# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""A pipeline to chain transformers.

The :class:`.Pipeline` class chains a sequence of tranformers, and provides global
fit(), transform(), fit_transform() and inverse_transform() methods.
"""
from __future__ import division, unicode_literals

from typing import Optional, Sequence

from numpy import eye, matmul, ndarray

from gemseo.mlearning.transform.transformer import Transformer, TransformerFitOptionType


class Pipeline(Transformer):
    """Transformer pipeline.

    Attributes:
        transformers (Sequence(Transformer)): The sequence of transformers.
    """

    def __init__(
        self,
        name="Pipeline",  # type: str
        transformers=None,  # type:Optional[Sequence[Transformer]]
    ):  # type: (...) -> None
        """
        Args:
            name: A name for this pipeline.
            transformers: A sequence of transformers to be
                chained. The transformers are chained in the order of appearance in
                the list, i.e. the first transformer is applied first. If
                transformers is an empty list or None, then the pipeline
                transformer behaves like an identity transformer.
        """
        super(Pipeline, self).__init__(name)
        self.transformers = transformers or []

    def duplicate(self):  # type: (...) -> Pipeline
        """Duplicate the current object.

        Returns:
            A deepcopy of the current instance.
        """
        transformers = [trans.duplicate() for trans in self.transformers]
        return self.__class__(self.name, transformers)

    def fit(
        self,
        data,  # type: ndarray
        **options  # type: TransformerFitOptionType
    ):  # type: (...) -> None
        """Fit the transformer pipeline to the data.

        All the transformers are fitted, transforming the data in place.

        Args:
            data: The data to be fitted.
        """
        for transformer in self.transformers:
            data = transformer.fit_transform(data, **options)

    def transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Transform the data.

        The data is transformed sequentially,
        where the output of one transformer is the input of the next.

        Args:
            data: The data to be transformed.

        Returns:
            The transformed data.
        """
        for transformer in self.transformers:
            data = transformer.transform(data)
        return data

    def inverse_transform(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Perform an inverse transform on the data.

        The data is inverse transformed sequentially,
        starting with the last transformer in the list.

        Args:
            data: The data to be inverse transformed.

        Returns:
            The inverse transformed data.
        """
        for transformer in self.transformers[::-1]:
            data = transformer.inverse_transform(data)
        return data

    def compute_jacobian(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the Jacobian of the ``pipeline.transform()``.

        Args:
            data: The data where the Jacobian is to be computed.

        Returns:
            The Jacobian matrix.
        """
        jacobian = eye(data.shape[-1])
        for transformer in self.transformers:
            jacobian = matmul(transformer.compute_jacobian(data), jacobian)
            data = transformer.transform(data)
        return jacobian

    def compute_jacobian_inverse(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        """Compute the Jacobian of the ``pipeline.inverse_transform()``.

        Args:
            data: The data where the Jacobian is to be computed.

        Returns:
            The Jacobian matrix.
        """
        jacobian = eye(data.shape[-1])
        for transformer in self.transformers[::-1]:
            jacobian = matmul(transformer.compute_jacobian_inverse(data), jacobian)
            data = transformer.inverse_transform(data)
        return jacobian



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""Draw a bar plot from a :class:`.Dataset`. """
from __future__ import division, unicode_literals

from typing import List, Mapping

import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from numpy import arange, linspace

from gemseo.post.dataset.dataset_plot import DatasetPlot, DatasetPlotPropertyType


class BarPlot(DatasetPlot):
    """Barplot visualization."""

    def _plot(
        self,
        properties,  # type: Mapping[str,DatasetPlotPropertyType]
        n_digits=1,  # type: int
    ):  # type: (...) -> List[Figure]
        """
        Args:
            n_digits: The number of digits to print the different bar values.
        """
        # radar solid grid lines
        all_data, _, sizes = self.dataset.get_all_data(False, False)
        variables_names = self.dataset.columns_names
        dimension = sum(sizes.values())
        series_names = self.dataset.row_names

        if self.color is None:
            colormap = plt.cm.get_cmap(self.colormap)
            self.color = {
                name: colormap(color)
                for name, color in zip(series_names, linspace(0, 1, len(all_data)))
            }

        fig, axe = plt.subplots()
        axe.tick_params(labelsize=self.font_size)

        discretization = arange(dimension)
        width = 0.75 / len(all_data)
        subplots = []
        positions = [
            discretization + index * width + width / 2 for index in range(dimension)
        ]
        for position, name, data in zip(positions, series_names, all_data):
            data = data.tolist()
            subplots.append(
                axe.bar(
                    position,
                    data,
                    width,
                    label=name,
                    color=self.color[name],
                )
            )

        for rects in subplots:
            for rect in rects:
                height = rect.get_height()
                if height > 0:
                    pos = 3
                else:
                    pos = -12
                axe.annotate(
                    "{}".format(round(height, n_digits)),
                    xy=(rect.get_x() + rect.get_width() / 2, height),
                    xytext=(0, pos),  # 3 points vertical offset
                    textcoords="offset points",
                    ha="center",
                    va="bottom",
                )

        axe.set_xticks(discretization)
        axe.set_xticklabels(variables_names)
        axe.set_title(self.title, fontsize=self.font_size * 1.2)
        axe.legend(fontsize=self.font_size)
        return [fig]



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""Computation of tolerance intervals from a data-fitted probability distribution."""
import logging
from typing import Any, Tuple, Type

import six
from custom_inherit import DocInheritMeta
from numpy import array, inf, ndarray

from gemseo.core.factory import Factory
from gemseo.utils.base_enum import BaseEnum

LOGGER = logging.getLogger(__name__)

ToleranceIntervalSide = BaseEnum("ToleranceIntervalSide", "LOWER UPPER BOTH")


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
        include_special_methods=True,
    )
)
class ToleranceInterval(object):
    """Computation of tolerance intervals from a data-fitted probability distribution.

    A :class:`.ToleranceInterval` (TI) is initialized
    from the number of samples
    used to estimate the parameters of the probability distribution
    and from the estimations of these parameters.

    A :class:`.ToleranceInterval` can be evaluated from:

    - a coverage defining the minimum percentage of belonging to the TI, e.g. 0.90,
    - a level of confidence in [0,1], e.g. 0.95,
    - a type of interval,
      either 'lower' for lower-sided TI,
      'upper' for upper-sided TI
      or 'both for both-sided TI.

    .. note::

       Lower-sided tolerance intervals are used to analyse the strength of materials.
       They are also known as *basis tolerance limits*.
       In particular,
       the *B-value* is the lower bound of the lower-sided tolerance interval
       with 90%-coverage and 95%-confidence
       while the *A-value* is the lower bound of the lower-sided tolerance interval
       with 95%-coverage and 95%-confidence.
    """

    def __init__(
        self,
        size,  # type: int
    ):  # type:(...) -> None
        # noqa: D205 D212 D415
        """
        Args:
            size: The number of samples.
        """
        self.__size = size

    def _compute_lower_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        """Compute the lower bound of the tolerance interval.

        Args:
            coverage: A minimum percentage of belonging to the TI.
            alpha: ``1-alpha`` is the level of confidence in [0,1].
            size: The number of samples.

        Returns:
            The lower bound of the tolerance interval.
        """
        raise NotImplementedError

    def _compute_upper_bound(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
    ):  # type: (...) -> float
        """Compute the upper bound of the tolerance interval.

        Args:
            coverage: A minimum percentage of belonging to the TI.
            alpha: ``1-alpha`` is the level of confidence in [0,1].
            size: The number of samples.

        Returns:
            The upper bound of the tolerance interval.
        """
        raise NotImplementedError

    def _compute(
        self,
        coverage,  # type: float
        alpha,  # type: float
        size,  # type: int
        side,  # type: ToleranceIntervalSide
    ):  # type: (...) -> Tuple[ndarray,ndarray]
        r"""Compute the bounds of the tolerance interval.

        Args:
            coverage: A minimum percentage of belonging to the TI.
            alpha: ``1-alpha`` is the level of confidence in [0,1].
            size: The number of samples.
            side: The type of the tolerance interval
                characterized by its *sides* of interest,
                either a lower-sided tolerance interval :math:`[a, +\infty[`,
                an upper-sided tolerance interval :math:`]-\infty, b]`,
                or a two-sided tolerance interval :math:`[c, d]`.

        Returns:
            The bounds of the tolerance interval.

        Raises:
            ValueError: If the type of tolerance interval is incorrect.
        """
        if side == ToleranceIntervalSide.LOWER:
            lower = self._compute_lower_bound(coverage, alpha, size)
            upper = inf
        elif side == ToleranceIntervalSide.UPPER:
            lower = -inf
            upper = self._compute_upper_bound(coverage, alpha, size)
        elif side == ToleranceIntervalSide.BOTH:
            coverage = (coverage + 1.0) / 2.0
            alpha = alpha / 2.0
            lower = self._compute_lower_bound(coverage, alpha, size)
            upper = self._compute_upper_bound(coverage, alpha, size)
        else:
            raise ValueError("The type of tolerance interval is incorrect.")
        return array([lower]), array([upper])

    def compute(
        self,
        coverage,  # type: float
        confidence=0.95,  # type: float
        side=ToleranceIntervalSide.BOTH,  # type: ToleranceIntervalSide
    ):  # type: (...) -> Tuple[ndarray,ndarray]
        r"""Compute a tolerance interval.

        Args:
            coverage: A minimum percentage of belonging to the TI.
            confidence: A level of confidence in [0,1].
            side: The type of the tolerance interval
                characterized by its *sides* of interest,
                either a lower-sided tolerance interval :math:`[a, +\infty[`,
                an upper-sided tolerance interval :math:`]-\infty, b]`,
                or a two-sided tolerance interval :math:`[c, d]`.

        Returns:
            The tolerance bounds.
        """
        return self._compute(coverage, 1 - confidence, self.__size, side)


class ToleranceIntervalFactory(object):
    """A factory of :class:`.ToleranceInterval`."""

    def __init__(self):  # type: (...) -> None
        # noqa: D107
        self.__factory = Factory(
            ToleranceInterval, ("gemseo.uncertainty.statistics.tolerance_interval",)
        )

    def create(
        self,
        class_name,  # type: str
        size,  # type: int
        *args  # type: float
    ):  # type: (...) -> ToleranceInterval
        """Return an instance of :class:`.ToleranceInterval`.

        Args:
            size: The number of samples
                used to estimate the parameters of the probability distribution.
            *args: The arguments of the probability distribution.

        Returns:
            The instance of the class.

        Raises:
            TypeError: If the class cannot be instantiated.
        """
        cls = self.get_class(class_name)
        try:
            return cls(size, *args)
        except TypeError:
            LOGGER.error(
                "Failed to create class %s with arguments %s", class_name, args
            )
            raise TypeError(
                "Cannot create {}ToleranceInterval with arguments {}".format(
                    class_name, args
                )
            )

    def get_class(
        self, name  # type: str
    ):  # type: (...) -> Type[Any]
        """Return a class from its name.

        Args:
            name: The name of the class.

        Returns:
            The class.
        """
        return self.__factory.get_class("{}ToleranceInterval".format(name))



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Variable space defining both deterministic and uncertain variables.

Overview
--------

The :class:`.ParameterSpace` class describes a set of parameters of interest
which can be either deterministic or uncertain.
This class inherits from :class:`.DesignSpace`.

Capabilities
------------

The :meth:`.DesignSpace.add_variable` aims to add deterministic variables from:

- a variable name,
- an optional variable size (default: 1),
- an optional variable type (default: float),
- an optional lower bound (default: - infinity),
- an optional upper bound (default: + infinity),
- an optional current value (default: None).

The :meth:`.add_random_variable` aims to add uncertain
variables (a.k.a. random variables) from:

- a variable name,
- a distribution name
  (see :meth:`~gemseo.uncertainty.api.get_available_distributions`),
- an optional variable size,
- optional distribution parameters (:code:`parameters` set as
  a tuple of positional arguments for :class:`.OTDistribution`
  or a dictionary of keyword arguments for :class:`.SPDistribution`,
  or keyword arguments for standard probability distribution such
  as :class:`.OTNormalDistribution` and :class:`.SPNormalDistribution`).

The :class:`.ParameterSpace` also provides the following methods:

- :meth:`.compute_samples`: returns several samples
  of the uncertain variables,
- :meth:`.evaluate_cdf`: evaluate the cumulative density function
  for the different variables and their different
- :meth:`.get_range` returns the numerical range
  of the different uncertain parameters,
- :meth:`.get_support`: returns the mathematical support
  of the different uncertain variables,
- :meth:`.is_uncertain`: checks if a parameter is uncertain,
- :meth:`.is_deterministic`: checks if a parameter is deterministic.
"""
from __future__ import division, unicode_literals

import logging
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Mapping, Optional, Union

if TYPE_CHECKING:
    from gemseo.core.dataset import Dataset

from numpy import array, ndarray

from gemseo.algos.design_space import DesignSpace
from gemseo.uncertainty.distributions.composed import ComposedDistribution
from gemseo.uncertainty.distributions.factory import (
    DistributionFactory,
    DistributionParametersType,
)
from gemseo.utils.data_conversion import DataConversion

LOGGER = logging.getLogger(__name__)


class ParameterSpace(DesignSpace):
    """Parameter space.

    Attributes:
        uncertain_variables (List(str)): The names of the uncertain variables.
        distributions (Dict(str,Distribution)): The marginal probability distributions
            of the uncertain variables.
        distribution (ComposedDistribution): The joint probability distribution
            of the uncertain variables.
    """

    _INITIAL_DISTRIBUTION = "Initial distribution"
    _TRANSFORMATION = "Transformation"
    _SUPPORT = "Support"
    _MEAN = "Mean"
    _STANDARD_DEVIATION = "Standard deviation"
    _RANGE = "Range"
    _BLANK = ""
    _PARAMETER_SPACE = "Parameter space"

    def __init__(
        self,
        copula=ComposedDistribution._INDEPENDENT_COPULA,  # type: str
    ):  # type: (...) -> None
        """
        Args:
            copula: A name of copula defining the dependency between random variables.
        """
        LOGGER.info("*** Create a new parameter space ***")
        super(ParameterSpace, self).__init__()
        self.uncertain_variables = []
        self.distributions = {}
        self.distribution = None
        if copula not in ComposedDistribution.AVAILABLE_COPULA_MODELS:
            raise ValueError("{} is not a copula name".format(copula))
        self._copula = copula

    def is_uncertain(
        self,
        variable,  # type: str
    ):  # type: (...) -> bool
        """Check if a variable is uncertain.

        Args:
            variable: The name of the variable.

        Returns:
            True is the variable is uncertain.
        """
        return variable in self.uncertain_variables

    def is_deterministic(
        self,
        variable,  # type: str
    ):  # type: (...) -> bool
        """Check if a variable is deterministic.

        Args:
            variable: The name of the variable.

        Returns:
            True is the variable is deterministic.
        """
        deterministic = set(self.variables_names) - set(self.uncertain_variables)
        return variable in deterministic

    def __update_parameter_space(
        self,
        variable,  # type: str
    ):  # type: (...) -> None
        """Update the parameter space with a random variable.

        Args:
            variable: The name of the random variable.
        """
        if variable not in self.variables_names:
            l_b = self.distributions[variable].math_lower_bound
            u_b = self.distributions[variable].math_upper_bound
            value = self.distributions[variable].mean
            size = self.distributions[variable].dimension
            self.add_variable(variable, size, "float", l_b, u_b, value)
        else:
            l_b = self.distributions[variable].math_lower_bound
            u_b = self.distributions[variable].math_upper_bound
            value = self.distributions[variable].mean
            self.set_lower_bound(variable, l_b)
            self.set_upper_bound(variable, u_b)
            self.set_current_variable(variable, value)

    def add_random_variable(
        self,
        name,  # type: str
        distribution,  # type: str
        size=1,  # type: int
        **parameters  # type: DistributionParametersType
    ):  # type: (...) -> None
        """Add a random variable from a probability distribution.

        Args:
            name: The name of the random variable.
            distribution: The name of a class
                implementing a probability distribution,
                e.g. 'OTUniformDistribution' or 'SPDistribution'.
            size: The dimension of the random variable.
            **parameters: The parameters of the distribution.
        """
        factory = DistributionFactory()
        distribution = factory.create(
            distribution, variable=name, dimension=size, **parameters
        )
        variable = distribution.variable_name
        LOGGER.info("Add the random variable: %s", variable)
        self.distributions[variable] = distribution
        self.uncertain_variables.append(variable)
        self._build_composed_distribution()
        self.__update_parameter_space(variable)

    def _build_composed_distribution(self):  # type: (...) -> None
        """Build the composed distribution from the marginal ones."""
        tmp_marginal = self.distributions[self.uncertain_variables[0]]
        marginals = [self.distributions[name] for name in self.uncertain_variables]
        self.distribution = tmp_marginal._COMPOSED_DISTRIBUTION(marginals, self._copula)

    def get_range(
        self,
        variable,  # type: str
    ):  # type: (...) -> List[ndarray]
        """Return the numerical range of a random variable.

        Args:
            variable: The name of the random variable.

        Returns:
            The range of the components of the random variable.
        """
        return self.distributions[variable].range

    def get_support(
        self,
        variable,  # type: str
    ):  # type: (...) -> List[ndarray]
        """Return the mathematical support of a random variable.

        Args:
            variable: The name of the random variable.

        Returns:
            The support of the components of the random variable.
        """
        return self.distributions[variable].support

    def remove_variable(
        self,
        name,  # type: str
    ):  # type: (...) -> None
        """Remove a variable from the probability space.

        Args:
            name: The name of the variable.
        """
        if name in self.uncertain_variables:
            del self.distributions[name]
            self.uncertain_variables.remove(name)
            if self.uncertain_variables:
                self._build_composed_distribution()
        super(ParameterSpace, self).remove_variable(name)

    def compute_samples(
        self,
        n_samples=1,  # type: int
        as_dict=False,  # type: bool
    ):  # type: (...) -> Union[Dict[str,ndarray],ndarray]
        """Sample the random variables and return the realizations.

        Args:
            n_samples: A number of samples.
            as_dict: The type of the returned object.
                If True, return a dictionary.
                Otherwise, return an array.

        Returns:
            The realizations of the random variables,
                either stored in an array or in a dictionary
                whose values are the names of the random variables
                and the values are the evaluations.
        """
        sample = self.distribution.compute_samples(n_samples)
        if as_dict:
            sample = [
                DataConversion.array_to_dict(
                    data_array, self.uncertain_variables, self.variables_sizes
                )
                for data_array in sample
            ]
        return sample

    def evaluate_cdf(
        self,
        value,  # type: Dict[str,ndarray]
        inverse=False,  # type:bool
    ):  # type: (...) -> Dict[str,ndarray]
        """Evaluate the cumulative density function (or its inverse) of each marginal.

        Args:
            value: The values of the uncertain variables
                passed as a dictionary whose keys are the names of the variables.
            inverse: The type of function to evaluate.
                If True, compute the cumulative density function.
                Otherwise, compute the inverse cumulative density function.

        Returns:
            A dictionary where the keys are the names of the random variables
                and the values are the evaluations.
        """
        if inverse:
            self.__check_dict_of_array(value)
        values = {}
        for name in self.uncertain_variables:
            val = value[name]
            distribution = self.distributions[name]
            if inverse:
                current_v = distribution.compute_inverse_cdf(val)
            else:
                current_v = distribution.compute_cdf(val)
            values[name] = array(current_v)

        return values

    def __check_dict_of_array(
        self,
        obj,  # type: Any
    ):  # type: (...) -> None
        """Check if the object is a dictionary whose values are numpy arrays.

        Args:
            obj: The object to test.
        """
        error_msg = (
            "obj must be a dictionary whose keys are the variables "
            "names and values are arrays whose dimensions are the "
            "variables ones and components are in [0, 1]"
        )
        if not isinstance(obj, dict):
            raise TypeError(error_msg)
        for variable, value in obj.items():
            if variable not in self.uncertain_variables:
                LOGGER.debug(
                    "%s is not defined in the probability space."
                    " Available variables are [%s]."
                    " Use uniform distribution for %s.",
                    variable,
                    ", ".join(self.uncertain_variables),
                    variable,
                )
            else:
                if not isinstance(value, ndarray):
                    raise TypeError(error_msg)
                if len(value.flatten()) != self.variables_sizes[variable]:
                    raise ValueError(error_msg)
                if any(value.flatten() > 1.0) or any(value.flatten() < 0.0):
                    raise ValueError(error_msg)

    def __str__(self):  # type: (...) -> str
        table = super(ParameterSpace, self).get_pretty_table()
        distribution = []
        for variable in self.variables_names:
            if variable in self.uncertain_variables:
                dist = self.distributions[variable]
                for _ in range(dist.dimension):
                    distribution.append(str(dist))
            else:
                for _ in range(self.variables_sizes[variable]):
                    distribution.append(self._BLANK)

        table.add_column(self._INITIAL_DISTRIBUTION, distribution)
        table.title = self._PARAMETER_SPACE
        desc = str(table)
        return desc

    def get_tabular_view(
        self,
        decimals=2,  # type: int
    ):  # type: (...) -> str
        """Return a tabular view of the parameter space.

        This view contains statistical information.

        Args:
            decimals: The number of decimals to print.

        Returns:
            The tabular view.
        """
        table = super(ParameterSpace, self).get_pretty_table()
        distribution = []
        transformation = []
        support = []
        mean = []
        std = []
        rnge = []
        for variable in self.variables_names:
            if variable in self.uncertain_variables:
                dist = self.distributions[variable]
                tmp_mean = dist.mean
                tmp_std = dist.standard_deviation
                tmp_range = dist.range
                tmp_support = dist.support
                for dim in range(dist.dimension):
                    distribution.append(str(dist))
                    transformation.append(dist.transformation)
                    mean.append(tmp_mean[dim])
                    mean[-1] = round(mean[-1], decimals)
                    std.append(tmp_std[dim])
                    std[-1] = round(std[-1], decimals)
                    rnge.append(tmp_range[dim])
                    support.append(tmp_support[dim])
            else:
                for _ in range(self.variables_sizes[variable]):
                    distribution.append(self._BLANK)
                    transformation.append(self._BLANK)
                    mean.append(self._BLANK)
                    std.append(self._BLANK)
                    support.append(self._BLANK)
                    rnge.append(self._BLANK)

        table.add_column(self._INITIAL_DISTRIBUTION, distribution)
        table.add_column(self._TRANSFORMATION, transformation)
        table.add_column(self._SUPPORT, support)
        table.add_column(self._MEAN, mean)
        table.add_column(self._STANDARD_DEVIATION, std)
        table.add_column(self._RANGE, rnge)
        table.title = self._PARAMETER_SPACE
        desc = str(table)
        return desc

    def unnormalize_vect(
        self,
        x_vect,  # type:ndarray
        minus_lb=True,  # type:bool
        no_check=False,  # type: bool
        use_dist=True,  # type:bool
    ):  # type: (...) ->ndarray
        """Unnormalize an unit vector.

        Args:
            x_vect: The unit vector to unnormalize.
            minus_lb: The type of normalization.
                If True, remove lower bounds at normalization.
            no_check: The data checker.
                If True, do not check that values are in [0,1].
            use_dist: The statistical scaling.
                If True, rescale wrt the statistical law.

        Returns:
            The unnormalized vector.
        """
        if not use_dist:
            return super(ParameterSpace, self).unnormalize_vect(x_vect)

        data_names = self.variables_names
        data_sizes = self.variables_sizes
        dict_sample = DataConversion.array_to_dict(x_vect, data_names, data_sizes)
        x_u_geom = super(ParameterSpace, self).unnormalize_vect(x_vect)
        x_u = self.evaluate_cdf(dict_sample, inverse=True)
        x_u_geom = DataConversion.array_to_dict(x_u_geom, data_names, data_sizes)
        missing_names = list(set(data_names) - set(x_u.keys()))
        for name in missing_names:
            x_u[name] = x_u_geom[name]
        x_u = DataConversion.dict_to_array(x_u, data_names)
        return x_u

    def normalize_vect(
        self,
        x_vect,  # type:ndarray
        minus_lb=True,  # type: bool
        use_dist=False,  # type: bool
    ):  # type: (...) ->ndarray
        """Normalize a vector.

        Args:
            x_vect: The vector to normalize.
            minus_lb: The type of normalization.
                If True, remove lower bounds at normalization.
            no_check: The data checker.
                If True, do not check that values are in [0,1].
            use_dist: The statistical scaling.
                If True, rescale wrt the statistical law.

        Returns:
            The normalized vector.
        """
        if use_dist:
            return super(ParameterSpace, self).normalize_vect(x_vect)

        data_names = self.variables_names
        data_sizes = self.variables_sizes
        dict_sample = DataConversion.array_to_dict(x_vect, data_names, data_sizes)
        x_u_geom = super(ParameterSpace, self).normalize_vect(x_vect)
        x_u = self.evaluate_cdf(dict_sample, inverse=False)
        x_u_geom = DataConversion.array_to_dict(x_u_geom, data_names, data_sizes)
        missing_names = list(set(data_names) - set(x_u.keys()))
        for name in missing_names:
            x_u[name] = x_u_geom[name]
        x_u = DataConversion.dict_to_array(x_u, data_names)
        return x_u

    @property
    def deterministic_variables(self):  # type: (...) -> List[str]
        """The deterministic variables."""
        return [
            variable
            for variable in self.variables_names
            if variable not in self.uncertain_variables
        ]

    def extract_uncertain_space(self):  # type: (...) -> ParameterSpace
        """Extract the uncertain space.

        Return:
            The uncertain space.
        """
        uncertain_space = deepcopy(self)
        uncertain_space.filter(self.uncertain_variables)
        return uncertain_space

    def extract_deterministic_space(self):  # type: (...) -> ParameterSpace
        """Extract the deterministic space.

        Return:
            The deterministic space.
        """
        deterministic_space = DesignSpace()
        for name in self.deterministic_variables:
            deterministic_space.add_variable(
                name, self.get_size(name), self.get_type(name)
            )
            value = self._current_x.get(name)
            if value is not None:
                deterministic_space.set_current_variable(name, value)
            deterministic_space.set_lower_bound(name, self.get_lower_bound(name))
            deterministic_space.set_upper_bound(name, self.get_upper_bound(name))
        return deterministic_space

    @staticmethod
    def init_from_dataset(
        dataset,  # type: Dataset
        groups=None,  # type: Optional[Iterable[str]]
        uncertain=None,  # type: Optional[Mapping[str,bool]]
        copula=ComposedDistribution._INDEPENDENT_COPULA,  # type: str
    ):  # type: (...) -> ParameterSpace
        """Initialize the parameter space from a dataset.

        Args:
            dataset: The dataset used for the initialization.
            groups: The groups of the dataset to be considered.
                If None, consider all the groups.
            uncertain: Whether the variables should be uncertain or not.
            copula: A name of copula defining the dependency between random variables.
        """
        parameter_space = ParameterSpace(copula=copula)

        if uncertain is None:
            uncertain = {}

        if groups is None:
            groups = dataset.groups
        for group in groups:
            for name in dataset.get_names(group):
                data = dataset.get_data_by_names(name)[name]
                l_b = data.min(0)
                u_b = data.max(0)
                value = (l_b + u_b) / 2
                size = len(l_b)

                if uncertain.get(name, False):
                    for idx in range(size):
                        parameter_space.add_random_variable(
                            "{}_{}".format(name, idx),
                            "OTUniformDistribution",
                            1,
                            minimum=float(l_b[idx]),
                            maximum=float(u_b[idx]),
                        )
                else:
                    parameter_space.add_variable(name, size, "float", l_b, u_b, value)

        return parameter_space



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

r"""Class for the estimation of Morris indices.

OAT technique
-------------

The purpose of the One-At-a-Time (OAT) methodology is to quantify the elementary effect

.. math::

   df_i = f(X_1+dX_1,\ldots,X_i+dX_i,\ldots,X_d)-f(X_1,\ldots,X_i,\ldots,X_d)

associated with a small variation :math:`dX_i` of :math:`X_i` with

.. math::

   df_1 = f(X_1+dX_1,\ldots,X_i,\ldots,X_d)-f(X_1,\ldots,X_i,\ldots,X_d)

The elementary effects :math:`df_1,\ldots,df_d` are computed sequentially
from an initial point

.. math::

   X=(X_1,\ldots,X_d)

From these elementary effects, we can compare their absolute values
:math:`|df_1|,\ldots,|df_d|` and sort :math:`X_1,\ldots,X_d` accordingly.

Morris technique
----------------

Then, the purpose of the Morris' methodology is to repeat the OAT method
from different initial points :math:`X^{(1)},\ldots,X^{(r)}`
and compare the parameters in terms of mean

.. math::

   \mu_i^* = \frac{1}{r}\sum_{j=1}^r|df_i^{(j)}|

and standard deviation

.. math::

   \sigma_i = \frac{1}{r}\sum_{j=1}^r\left(|df_i^{(j)}|-\mu_i\right)^2

where :math:`\mu_i = \frac{1}{r}\sum_{j=1}^rdf_i^{(j)}`.

This methodology relies on the :class:`.MorrisAnalysis` class.
"""

from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Mapping, Optional, Sequence, Tuple, Union

import matplotlib.pyplot as plt
from numpy import abs as np_abs
from numpy import array

from gemseo.algos.design_space import DesignSpace
from gemseo.algos.doe.lib_pydoe import PyDOE
from gemseo.core.discipline import MDODiscipline
from gemseo.uncertainty.sensitivity.analysis import IndicesType, SensitivityAnalysis
from gemseo.uncertainty.sensitivity.morris.oat import OATSensitivity
from gemseo.utils.py23_compat import Path

LOGGER = logging.getLogger(__name__)


class MorrisAnalysis(SensitivityAnalysis):
    r"""Sensitivity analysis based on the Morris' indices.

    :attr:`.MorrisAnalysis.indices` contains both :math:`\mu^*`, :math:`\mu`
    and :math:`\sigma` while :attr:`.MorrisAnalysis.main_indices`
    represents :math:`\mu^*`. Lastly, the :meth:`.MorrisAnalysis.plot`
    method represents the parameters as a scatter plot
    where :math:`X_i` has as coordinates :math:`(\mu_i^*,\sigma_i)`.
    The bigger :math:`\mu_i^*` is, the more significant :math:`X_i` is.
    Concerning :math:`\sigma_i`, it highlights non-linear effects
    along :math:`X_i` or cross-effects between :math:`X_i` and other parameter(s).

    The user can specify the DOE algorithm name to select the initial points, as
    well as the number of replicates and the relative step for the input variations.

    Attributes:
        mu_ (dict): The mean effects with the following structure:

            .. code-block:: python

                {
                    "output_name": [
                        {
                            "input_name": data_array,
                        }
                    ]
                }

        mu_star (dict): The mean absolute effects with the following structure:

            .. code-block:: python

                {
                    "output_name": [
                        {
                            "input_name": data_array,
                        }
                    ]
                }

        sigma (dict): The variability of the effects with the following structure:

            .. code-block:: python

                {
                    "output_name": [
                        {
                            "input_name": data_array,
                        }
                    ]
                }

    Examples:
        >>> from numpy import pi
        >>> from gemseo.api import create_discipline, create_parameter_space
        >>> from gemseo.uncertainty.sensitivity.morris.analysis import MorrisAnalysis
        >>>
        >>> expressions = {"y": "sin(x1)+7*sin(x2)**2+0.1*x3**4*sin(x1)"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x3", "OTUniformDistribution", minimum=-pi, maximum=pi
        ... )
        >>>
        >>> analysis = MorrisAnalysis(discipline, parameter_space, n_replicates=5)
        >>> indices = analysis.compute_indices()
    """

    DEFAULT_DRIVER = PyDOE.PYDOE_LHS

    def __init__(
        self,
        discipline,  # type: MDODiscipline
        parameter_space,  # type:DesignSpace
        n_samples,  # type: int
        algo=None,  # type: Optional[str]
        algo_options=None,  # type: Optional[Mapping]
        n_replicates=5,  # type: int
        step=0.05,  # type: float
    ):  # type: (...) -> None
        # noqa: D205,D212,D415
        r"""
        Args:
            n_replicates (int, optional): The number of times
                the OAT method is repeated. Used only if :attr:`n_samples` is None.
                Otherwise, this number is the greater integer :math:`r`
                such that :math:`r(d+1)\leq` :attr:`n_samples`
                and :math:`r(d+1)` is the number of samples actually carried out.
            step (float, optional): The finite difference step of the OAT method.
        """
        self.mu_ = None
        self.mu_star = None
        self.sigma = None
        self.__step = step
        if n_samples is None:
            self.__n_replicates = n_replicates
        else:
            self.__n_replicates = n_samples // (parameter_space.dimension + 1)
        self.__outputs = discipline.get_output_data_names()

        if parameter_space.dimension != len(parameter_space.variables_names):
            raise ValueError("Each input dimension must be equal to 1.")

        self.__diff_discipline = OATSensitivity(discipline, parameter_space, step)
        super(MorrisAnalysis, self).__init__(
            self.__diff_discipline, parameter_space, n_replicates, algo, algo_options
        )
        self._main_method = "Morris(mu*)"
        self.default_output = list(discipline.get_output_data_names())

    @property
    def n_replicates(self):  # type: (...) -> int
        """The number of OAT replicates."""
        return self.__n_replicates

    def compute_indices(
        self, outputs=None  # type: Optional[Sequence[str]]
    ):  # type: (...) -> Dict[str,IndicesType]
        # noqa: D102
        fd_data = self.dataset.get_data_by_group(self.dataset.OUTPUT_GROUP, True)
        output_names = outputs or self.default_output
        if not isinstance(output_names, list):
            output_names = [output_names]
        self.mu_ = {name: {} for name in output_names}
        self.mu_star = {name: {} for name in output_names}
        self.sigma = {name: {} for name in output_names}
        for fd_name, value in fd_data.items():
            output_name, input_name = self.__diff_discipline.get_io_names(fd_name)
            if output_name in output_names:
                self.mu_[output_name][input_name] = value.mean(0)
                self.mu_star[output_name][input_name] = np_abs(value).mean(0)
                self.sigma[output_name][input_name] = value.std(0)
        for output_name in output_names:
            length = len(next(iter(self.sigma[output_name].values())))
            self.mu_[output_name] = [
                {name: array([val[idx]]) for name, val in self.mu_[output_name].items()}
                for idx in range(length)
            ]
            self.mu_star[output_name] = [
                {
                    name: array([val[idx]])
                    for name, val in self.mu_star[output_name].items()
                }
                for idx in range(length)
            ]
            self.sigma[output_name] = [
                {
                    name: array([val[idx]])
                    for name, val in self.sigma[output_name].items()
                }
                for idx in range(length)
            ]
        return self.indices

    @property
    def indices(
        self,
    ):  # type: (...) -> IndicesType # noqa: D102
        return {"mu": self.mu_, "mu_star": self.mu_star, "sigma": self.sigma}

    @property
    def main_indices(
        self,
    ):  # type: (...) -> IndicesType # noqa: D102
        return self.mu_star

    def plot(
        self,
        output,  # type: Union[str,Tuple[str,int]]
        inputs=None,  # type: Optional[Iterable[str]]
        title=None,  # type: Optional[str]
        save=True,  # type: bool
        show=False,  # type: bool
        file_path=None,  # type: Optional[Union[str,Path]]
        directory_path=None,  # type: Optional[Union[str,Path]]
        file_name=None,  # type: Optional[str]
        file_format=None,  # type: Optional[str]
        offset=1,  # type: float
        lower_mu=None,  # type: Optional[float]
        lower_sigma=None,  # type: Optional[float]
    ):  # type: (...) -> None # noqa: D417
        r"""Plot the Morris indices for each input variable.

        For :math:`i\in\{1,\ldots,d\}`,
        plot :math:`\mu_i^*` in function of :math:`\sigma_i`.

        Args:
            offset: The offset to display the inputs names,
                expressed as a percentage applied to both x-range and y-range.
            lower_mu: The lower bound for :math:`\mu`.
                If None, use a default value.
            lower_sigma: The lower bound for :math:`\sigma`.
                If None, use a default value.
        """
        if not isinstance(output, tuple):
            output = (output, 0)
        names = self.dataset.get_names(self.dataset.INPUT_GROUP)
        names = self._filter_names(names, inputs)
        x_val = [self.mu_star[output[0]][output[1]][name] for name in names]
        y_val = [self.sigma[output[0]][output[1]][name] for name in names]
        fig, ax = plt.subplots()
        ax.scatter(x_val, y_val)
        ax.set_xlabel(r"$\mu^*$")
        ax.set_ylabel(r"$\sigma$")
        output = "{}({})".format(output[0], output[1])
        default_title = "Sampling: {}(size={}) - Relative step: {} - Output: {}"
        default_title = default_title.format(
            self._algo_name, self.__n_replicates, self.__step, output
        )
        ax.set_xlim(left=lower_mu)
        ax.set_ylim(bottom=lower_sigma)
        ax.set_title(title or default_title)
        x_offset = offset * (max(x_val) - min(x_val)) / 100.0
        y_offset = offset * (max(y_val) - min(y_val)) / 100.0
        for index, txt in enumerate(names):
            ax.annotate(txt, (x_val[index] + x_offset, y_val[index] + y_offset))
        self._save_show_plot(
            fig,
            save=save,
            show=show,
            file_path=file_path,
            file_name=file_name,
            file_format=file_format,
            directory_path=directory_path,
        )



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                        documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
SSBJ Disciplines wrappers
*************************
"""

from __future__ import division, unicode_literals

import time
from numbers import Number

from gemseo.core.discipline import MDODiscipline
from gemseo.problems.sobieski.core import SobieskiProblem

DTYPE_COMPLEX = "complex128"
DTYPE_DOUBLE = "float64"


class SobieskiBaseWrapper(MDODiscipline):
    """Base wrapper for Sobieski problem discipline wrappers and JSON grammars."""

    def __init__(self, dtype=DTYPE_DOUBLE):
        """Constructor.

        :param dtype: type of data, either "float64" or "complex128".
        :type dtype: str
        """
        super(SobieskiBaseWrapper, self).__init__(auto_detect_grammar_files=True)
        self.dtype = dtype
        self.sobieski_problem = SobieskiProblem(dtype=dtype)
        self.default_inputs = self.sobieski_problem.get_default_inputs(
            self.get_input_data_names()
        )
        self.re_exec_policy = self.RE_EXECUTE_DONE_POLICY

    def get_attributes_to_serialize(self):
        """Defines the attributes to be serialized Can be overloaded by disciplines.

        :returns: the list of attributes names
        """
        base_d = super(SobieskiBaseWrapper, self).get_attributes_to_serialize()
        base_d.append("dtype")
        return base_d

    def __setstate__(self, d):
        """Used by pickle to define what to deserialize.

        :param d : update self dict from d to deserialize
        """
        super(SobieskiBaseWrapper, self).__setstate__(d)
        self.sobieski_problem = SobieskiProblem(self.dtype)

    def _run(self):
        """Run the discipline."""
        raise NotImplementedError()


class SobieskiMission(SobieskiBaseWrapper):

    """Sobieski range wrapper using Breguet formula."""

    def __init__(self, dtype=DTYPE_DOUBLE, enable_delay=False):
        """Constructor of wrapper for range computation.

        :param dtype: type of data, either "float64" or "complex128".
        :type dtype: str
        """
        super(SobieskiMission, self).__init__(dtype=dtype)
        self.enable_delay = enable_delay

    def _run(self):
        """Compute range."""
        if self.enable_delay:
            if isinstance(self.enable_delay, Number):
                time.sleep(self.enable_delay)
            else:
                time.sleep(1.0)

        data_names = ["y_14", "y_24", "y_34", "x_shared"]
        y_14, y_24, y_34, x_shared = self.get_inputs_by_name(data_names)
        y_4 = self.sobieski_problem.blackbox_mission(x_shared, y_14, y_24, y_34)
        self.store_local_data(y_4=y_4)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the partial derivatives of all outputs wrt all inputs.

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["y_14", "y_24", "y_34", "x_shared"]
        y_14, y_24, y_34, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_mission(
            x_shared, y_14, y_24, y_34
        )


class SobieskiStructure(SobieskiBaseWrapper):

    """Sobieski mass estimation wrapper."""

    def __init__(self, dtype=DTYPE_DOUBLE):
        """Constructor of wrapper for weight computation.

        :param dtype: type of data, either "float64" or "complex128".
        :type dtype: str
        """
        super(SobieskiStructure, self).__init__(dtype=dtype)

    def _run(self):
        """Compute weight."""

        data_names = ["x_1", "y_21", "y_31", "x_shared"]
        x_1, y_21, y_31, x_shared = self.get_inputs_by_name(data_names)
        y_1, y_11, y_12, y_14, g_1 = self.sobieski_problem.blackbox_structure(
            x_shared, y_21, y_31, x_1
        )
        self.store_local_data(y_1=y_1, y_11=y_11, y_12=y_12, y_14=y_14, g_1=g_1)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Linearization of weight analysis.

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["x_1", "y_21", "y_31", "x_shared"]
        x_1, y_21, y_31, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_structure(
            x_shared, y_21, y_31, x_1
        )


class SobieskiAerodynamics(SobieskiBaseWrapper):

    """Sobieski aerodynamic discipline wrapper."""

    def __init__(self, dtype=DTYPE_DOUBLE):
        """Constructor of wrapper for aerodynamic computation.

        :param dtype: type of data, "float64" or "complex128".
        :type dtype: str
        """
        super(SobieskiAerodynamics, self).__init__(dtype=dtype)

    def _run(self):
        """Compute aerodynamics."""
        data_names = ["x_2", "y_12", "y_32", "x_shared"]
        x_2, y_12, y_32, x_shared = self.get_inputs_by_name(data_names)
        y_2, y_21, y_23, y_24, g_2 = self.sobieski_problem.blackbox_aerodynamics(
            x_shared, y_12, y_32, x_2
        )
        self.store_local_data(y_2=y_2, y_21=y_21, y_23=y_23, y_24=y_24, g_2=g_2)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the partial derivatives of all outputs wrt all inputs.

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["x_2", "y_12", "y_32", "x_shared"]
        x_2, y_12, y_32, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_aerodynamics(
            x_shared, y_12, y_32, x_2
        )


class SobieskiPropulsion(SobieskiBaseWrapper):
    """Sobieski propulsion propulsion wrapper."""

    def __init__(self, dtype=DTYPE_DOUBLE):
        """Constructor of wrapper for propulsion computation.

        :param dtype: type of data, either "float64" or "complex128"
        :type dtype: str
        """
        super(SobieskiPropulsion, self).__init__(dtype=dtype)

    def _run(self):
        """Compute propulsion."""
        data_names = ["x_3", "y_23", "x_shared"]

        x_3, y_23, x_shared = self.get_inputs_by_name(data_names)
        y_3, y_34, y_31, y_32, g_3 = self.sobieski_problem.blackbox_propulsion(
            x_shared, y_23, x_3
        )
        self.store_local_data(y_3=y_3, y_34=y_34, y_31=y_31, y_32=y_32, g_3=g_3)

    def _compute_jacobian(self, inputs=None, outputs=None):
        """Compute the partial derivatives of all outputs wrt all inputs.

        :param inputs: Default value = None)
        :param outputs: Default value = None)
        """
        data_names = ["x_3", "y_23", "x_shared"]
        x_3, y_23, x_shared = self.get_inputs_by_name(data_names)
        self.jac = self.sobieski_problem.derive_blackbox_propulsion(x_shared, y_23, x_3)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Make a discipline from an executable
************************************
"""
from __future__ import division, unicode_literals

import logging
import re
import subprocess
import sys
from ast import literal_eval
from copy import deepcopy
from multiprocessing import Lock, Manager
from os import listdir, mkdir
from os.path import join
from typing import Dict, List, Mapping, Optional, Sequence, Tuple, Union
from uuid import uuid1

from numpy import array, ndarray

from gemseo.core.data_processor import DataProcessor  # noqa: F401
from gemseo.core.data_processor import FloatDataProcessor
from gemseo.core.discipline import MDODiscipline
from gemseo.utils.base_enum import BaseEnum
from gemseo.utils.py23_compat import OrderedDict  # automatically dict from py36
from gemseo.utils.py23_compat import Path, xrange

LOGGER = logging.getLogger(__name__)

NUMERICS = [str(j) for j in xrange(10)]
INPUT_REGEX = r"GEMSEO_INPUT\{(.*)\}"
OUTPUT_REGEX = r"GEMSEO_OUTPUT\{(.*)\}"


class FoldersIter(BaseEnum):
    NUMBERED = 0
    UUID = 1


class Parsers(BaseEnum):
    KEY_VALUE_PARSER = 0
    TEMPLATE_PARSER = 1
    CUSTOM_CALLABLE = 2


class DiscFromExe(MDODiscipline):
    """Generic wrapper for executables.

    The DiscFromExe is a generic wrapper for executables. It generates a
    MDODiscipline from an executable and in inputs/output files wrappers.  The
    input and output files are described by templates.  The templates can be
    generated by executing the module
    :mod:`~gemseo.wrappers.template_grammar_editor` to open a GUI.


     It requires the creation of templates for input and output file,
     for instance, from the following input JSON file:

     .. code::

        {
        "a": 1.01515112125,
        "b": 2.00151511213,
        "c": 3.00151511213
        }

     A template that declares the inputs must be generated under this format,
     where "a" is the name of the input, and "1.0" is the default input.
     GEMSEO_INPUT declares an input, GEMSEO_OUTPUT declares an output, similarly.

     .. code::

        {
        "a": GEMSEO_INPUT{a::1.0},
        "b": GEMSEO_INPUT{b::2.0},
        "c": GEMSEO_INPUT{c::3.0}
        }

    Current limitations :

    - Only one input and one output file, otherwise,
     inherit from this class and modify the parsers.
     Only limited input writing and output parser strategies
     are implemented. To change that, you can pass custom parsing and
     writing methods to the constructor.
    - The only limitation in the current file format is that
     it must be a plain text file and not a binary file.
     In this case, the way of interfacing it is
     to provide a specific parser to the DiscFromExe,
     with the write_input_file_method
     and parse_outfile_method arguments of the constructor.

    Attributes:
        input_template (str): The path to the input template file.
        ouput_template (str): The path to the output template file.
        input_filename (str): The name of the input file.
        output_filename (str): The name of the ouput file.
        executable_command (str): The executable command.
        parse_outfile (Callable[Mapping[str, Tuple[int]], Sequence[str]]):
            The function used to parse the output file.
        write_input_file
            (Callable[str, Mapping[str, ndarray], Mapping[str, Tuple[int]], Sequence[int], str):
            The function used to write the input file.
        folder_iter (str): The method to be used to name new execution directories.
        output_folder_basepath (str): The base path of the execution directories.
        data_processor (DataProcessor): A data processor to be used before the execution
            of the discipline.
    """

    def __init__(
        self,
        input_template,  # type: str
        output_template,  # type: str
        output_folder_basepath,  # type: str
        executable_command,  # type: str
        input_filename,  # type: str
        output_filename,  # type: str
        folders_iter=FoldersIter.NUMBERED,  # type: Union[str, FoldersIter]
        name=None,  # type: Optional[str]
        parse_outfile_method=Parsers.TEMPLATE_PARSER,  # type: Union[str, Parsers]
        write_input_file_method=None,  # type: Optional[str]
        parse_out_separator="=",  # type: str
        use_shell=True,  # type: bool
    ):  # type: (...) -> None
        """
        Args:
            input_template: The path to the input file template.
                The input locations in the file are marked
                by GEMSEO_INPUT{input_name::1.0},
                where "input_name" is the input name, and 1.0 is here
                the default input.
            output_template: The path to the output file template.
                The input locations in the file are marked
                by GEMSEO_OUTPUT{output_name::1.0},
                where "output_name" is the input name.
            output_folcer_basepath: The path to the output folder,
                in which the executions will be performed.
            executable_command: The command to run the executable.
                Will be called through a system call.
                Example: "python myscript.py -i input.txt -o output.txt
            input_filename: The name of the input file.
                This will determine the name
                of the input file generated in the output folder.
                Example "input.txt".
            output_filename: The name of the output file.
                This will determine the name
                of the output file generated in the output folder.
                Example "output.txt".
            folders_iter (Union[str, FoldersIter]: The type of unique identifiers
                for the output folders. If NUMBERED the generated output folders
                will be "output_folder_basepath"+str(i+1),
                where i is the maximum value of the already existing
                "output_folder_basepath"+str(i) folders.
                Otherwise, a unique number based on the UUID function is
                generated. This last option shall be used if multiple MDO
                processes are runned in the same work directory.
            name: the name of the discipline. If None,
                use the class name.
            parse_outfile_method: The optional method that can be provided
                by the user to parse the output file. To see the signature of
                the method, see the parse_outfile method of this file.
                If the KEY_VALUE_PARSER is used as
                output parser, specify the separator key (default : "=").
            write_input_file_method: The method to write the input file.
                If None, use this modules' write_input_file. To see the signature
                of the method, see the write_input_file method of this file.
            parse_out_separator: The separator used for the output parser.
            use_shell: If True, run the command using the default shell. Otherwise,
                run directly the command.

        Raises:
            TypeError: If the provided write_input_file_method is not callable.
        """
        super(DiscFromExe, self).__init__(name=name)
        self.input_template = input_template
        self.output_template = output_template
        self.input_filename = input_filename
        self.output_filename = output_filename
        self.executable_command = executable_command
        self.__use_shell = use_shell

        use_template_parse = parse_outfile_method == Parsers.TEMPLATE_PARSER
        if parse_outfile_method is None or use_template_parse:
            self.parse_outfile = parse_outfile
            self._parse_outfile_method = Parsers.TEMPLATE_PARSER
        elif parse_outfile_method == Parsers.KEY_VALUE_PARSER:
            self.parse_outfile = lambda a, b: parse_key_value_file(
                a, b, parse_out_separator
            )
            self._parse_outfile_method = Parsers.KEY_VALUE_PARSER
        else:
            self.parse_outfile = parse_outfile_method
            self._parse_outfile_method = Parsers.CUSTOM_CALLABLE

        if not callable(self.parse_outfile):
            raise TypeError("The parse_outfile_method must be callable")

        self.write_input_file = write_input_file_method or write_input_file

        if not callable(self.write_input_file):
            raise TypeError("The write_input_file_method must be callable")

        self.__lock = Lock()

        self.__folders_iter = None
        self.folders_iter = folders_iter

        self.output_folder_basepath = output_folder_basepath

        self.__check_basepath_on_windows()

        self._out_pos = None
        self._in_dict = None
        self._out_dict = None
        self._in_lines = None
        self._out_lines = None

        n_dirs = self._get_max_outdir()
        self._counter = Manager().Value("i", n_dirs)

        self.data_processor = FloatDataProcessor()

        self.__parse_templates_and_set_grammars()

    @property
    def folders_iter(self):  # type: (...) -> FoldersIter
        """Getter/Setter for folders_iter.

        The setter will check that the value provided for folder_iter is valid.
        This check is done by checking its presence in FOLDERS_ITER.

        Raises:
            ValueError: If the value provided to the setter is not present
                in the accepted list of folders_iters list.
        """
        return self.__folders_iter

    @folders_iter.setter
    def folders_iter(
        self,
        value,  # type: Union[str, FoldersIter]
    ):  # type: (...) -> None
        if value not in FoldersIter:
            msg = "{} is not a valid folder_iter value.".format(value)
            raise ValueError(msg)
        self.__folders_iter = FoldersIter.get_member_from_name(value)

    def __check_basepath_on_windows(self):  # type: (...) -> None
        """Check that the basepath can be used.

        If the user use shell=True under Windows with a basepath
            that is on a network location, raise an error.

        Raises:
            ValueError: The basepath is located on a network location
                and cannot be run with cmd.exe.
        """
        if sys.platform.startswith("win") and self.__use_shell:
            resolved_basepath = Path(self.output_folder_basepath).resolve()
            if not resolved_basepath.parts[0].startswith("\\\\"):
                return

            msg = (
                "A network basepath and use_shell cannot be used together"
                " under Windows, as cmd.exe cannot change the current directory"
                " to a UNC path."
                " Please try use_shell=False or use a local base path."
            )
            raise ValueError(msg)

    def __parse_templates_and_set_grammars(self):  # type: (...) -> None
        """Parse the templates and set the grammar of the discipline."""
        with open(self.input_template, "r") as infile:
            self._in_lines = infile.readlines()
        with open(self.output_template, "r") as outfile:
            self._out_lines = outfile.readlines()

        self._in_dict, self._in_pos = parse_template(self._in_lines, True)
        self.input_grammar.initialize_from_data_names(self._in_dict.keys())

        out_dict, self._out_pos = parse_template(self._out_lines, False)

        self.output_grammar.initialize_from_data_names(out_dict.keys())

        msg = "Initialize discipline from template. \
                Input grammar: {}".format(
            self._in_dict.keys()
        )
        LOGGER.debug(msg)
        msg = "Initialize discipline from template. \
                Output grammar: {}".format(
            out_dict.keys()
        )
        LOGGER.debug(msg)
        self.default_inputs = {
            k: array([literal_eval(v)]) for k, v in self._in_dict.items()
        }

    def _run(self):  # type: (...) -> None
        """Run the wrapper."""
        uuid = self.generate_uid()

        out_dir = join(self.output_folder_basepath, uuid)

        mkdir(out_dir)
        input_file_path = join(out_dir, self.input_filename)

        self.write_input_file(
            input_file_path, self.local_data, self._in_pos, self._in_lines
        )

        if self.__use_shell:
            executable_command = self.executable_command
        else:
            executable_command = self.executable_command.split()

        err = subprocess.call(
            executable_command,
            shell=self.__use_shell,
            stderr=subprocess.STDOUT,
            cwd=out_dir,
        )
        if err != 0:
            raise RuntimeError("Execution failed and returned error code : " + str(err))
        outfile = join(out_dir, self.output_filename)
        with open(outfile, "r") as outfile:
            out_lines = outfile.readlines()

        if len(out_lines) != len(self._out_lines):
            raise ValueError(
                "The number of lines of the output file changed."
                "This is not supported yet"
            )

        out_vals = self.parse_outfile(self._out_pos, out_lines)
        self.local_data.update(out_vals)

    def generate_uid(self):  # type: (...) -> str
        """Generate an unique identifier for the execution directory.

        Generate a unique identifier for the current execution.
        If the folders_iter strategy is NUMBERED,
        the successive iterations are named by an integer 1, 2, 3 etc.
        This is multiprocess safe.
        Otherwise, a unique number based on the UUID function is generated.
        This last option shall be used if multiple MDO processes are runned
        in the same workdir.

        Returns:
            An unique string identifier (either a number or a UUID).
        """
        if self.folders_iter == FoldersIter.NUMBERED:
            with self.__lock:
                self._counter.value += 1
                return str(self._counter.value)
        elif self.folders_iter == FoldersIter.UUID:
            return str(uuid1()).split("-")[-1]
        else:
            msg = (
                "{} is not a valid method for creating the execution"
                " directories.".format(self.folders_iter)
            )
            raise ValueError(msg)

    def _list_out_dirs(self):  # type: (...) -> List[str]
        """Return the directories in the output folder path.

        Returns:
             The list of the directories in the output folder path.
        """
        return listdir(self.output_folder_basepath)

    def _get_max_outdir(self):  # type: (...) -> int
        """Get the maximum current index of output folders.

        Returns:
             The maximum index in the output folders.
        """
        outs = list(self._list_out_dirs())
        if not outs:
            return 0
        return max([literal_eval(n) for n in outs])


def parse_template(
    template_lines,  # type: Sequence[str]
    grammar_is_input,  # type: bool
):  # type: (...) -> Tuple[Dict[str, ndarray], Dict[str, Tuple[int]]]
    """Parse the input or output template.

    Args:
        template_lines: The lines of the template file.
        grammar_is_input: True for an input template, False otherwise.

    This function parses the input (or output) template.
    It returns the tuple (data_dict, pos_dict), where:

    - `data_dict` is the `{name:value}` dict:
       - name is the data name
       - value is the parsed input or output value in the template
    - `pos_dict` describes the template format {data_name:(start,end,line_number)}:
       - `data_name` is the name of the input data
       - `start` is the index of the starting point in the input file template.
        This index is a line index (character number on the line)
       - `end` is the index of the end character in the template
       - `line_number` is the index of the line in the file

    Returns:
        A data structure containing the parsed inpout or output template.
    """
    pattern_re = INPUT_REGEX if grammar_is_input else OUTPUT_REGEX

    regex = re.compile(pattern_re)  # , re.MULTILINE
    data_dict = OrderedDict()
    pos_dict = OrderedDict()

    for lineid, line in enumerate(template_lines):
        for match in regex.finditer(line):
            data = match.groups()[0]
            spl = data.split("::")
            name = spl[0]
            val = spl[1]
            data_dict[name] = val
            # When input mode: erase the template value
            if grammar_is_input:
                start, end = match.start(), match.end()
            else:
                # In output mode : catch all
                # the output lenght and not more
                start = match.start()
                end = start + len(val)

            pos_dict[name] = (start, end, lineid)

    return data_dict, pos_dict


def write_input_file(
    input_file_path,  # type: str
    data,  # type: Mapping[str, ndarray]
    input_positions,  # type: Mapping[str, Tuple[int]]
    input_lines,  # type: Sequence[int]
    float_format="{:1.18g}",  # type: str
):  # type: (...) -> None
    """Write the input file from the input data.

    Args:
        input_file_path: The absolute path to the file to be written.
        data: The local data of the discipline.
        input_positions: The information from the template
         format {data_name:(start,end,line_number)}, where name is the name of the input data,
         start is the index of the starting point in the input file template.
         This index is a line index (character number on the line).
         end is the index of the end character in the template,
         line_number is the index of the line in the file.
        input_lines: The lines of the input file template.
        float_format: The formating of the input data in the file (Default value = "{:1.18g}").
    """
    f_text = deepcopy(input_lines)
    for name, pos in input_positions.items():
        start, end, lineid = pos
        data_str = float_format.format(data[name])
        cline = f_text[lineid]
        f_text[lineid] = cline[:start] + data_str + cline[end:]

    with open(input_file_path, "w") as infile_o:
        infile_o.writelines(f_text)


def parse_key_value_file(
    _,
    out_lines,  # type: Sequence[str]
    separator="=",  # type: str
):  # type: (...) -> Dict[str, ndarray]
    """Parse the output file from the expected text positions.

    Args:
        out_lines: The lines of the output file template.
        separator: The separating characters of the key=value format.

    Returns:
        The output data in `.MDODiscipline` friendly data structure (e.g. Dict[str, ndarray]).
    """
    data = {}
    for line in out_lines:
        if separator in line:
            spl = line.strip().split(separator)
            if len(spl) != 2:
                raise ValueError("unbalanced = in line " + str(line))
            key = spl[0].strip()
            try:
                data[key] = float(literal_eval(spl[1].strip()))
            except Exception:
                raise ValueError("Failed to parse value as float " + str(spl[1]))

    return data


def parse_outfile(
    output_positions,  # type: Mapping[str, Tuple[int]]
    out_lines,  # type: Sequence[str]
):  # type: (...) -> Dict[str, ndarray]
    """Parse the output file from the expected text positions.

    Args:
        output_positions: The output position for each data name.
         The information from the template format
         {data_name:(start,end,dictionary)}, where
         name is the name of the output
         data, start is the index of the starting point
         in the input file template.
         This index is a line index (character number on the line)
         end is the index of the end character in the template
         line_number is the index of the line in the file
        out_lines: The lines of the output file template.

    Returns:
        The output data in `.MDODiscipline` friendly data structure (e.g. Dict[str, ndarray]).
    """
    values = {}
    for name, pos in output_positions.items():
        start, _, lineid = pos

        found_dot = False
        found_e = False
        # In case generated files has less lines
        if lineid > len(out_lines) - 1:
            break
        out_text = out_lines[lineid]
        i = start
        maxi = len(out_text)
        while True:
            # The problem is that the output file used for the template may be
            # using a output that is longer or shorter than the one generated
            # at runtime. We must find the proper end of the expression...
            i += 1
            char = out_text[i]
            if char == ".":
                # We found the . in float notation
                if found_dot or found_e:
                    break
                found_dot = True
                continue
            # We found the e in exp notation
            if char in ("E", "e"):
                if found_e:
                    break
                found_e = True
                continue
            # Check that we have nout reached EOL or space or whatever
            if char not in NUMERICS:
                break
            if i == maxi - 1:
                break
        outv = out_text[pos[0] : i]

        LOGGER.info("Parsed %s got output %s", name, outv)
        values[name] = array([float(outv)])
    return values



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Various termination criteria for drivers
****************************************
"""

from __future__ import division, unicode_literals

from numpy import all, allclose, average


class TerminationCriterion(Exception):
    """Stop driver for some reason."""


class FunctionIsNan(TerminationCriterion):
    """Stops driver when a function has NaN value or NaN Jacobian."""


class DesvarIsNan(TerminationCriterion):
    """Stops driver when the design variables are nan."""


class MaxIterReachedException(TerminationCriterion):
    """Exception raised when the maximum number of iterations is reached by the
    driver."""


class MaxTimeReached(TerminationCriterion):
    """Exception raised when the maximum execution time is reached by the driver."""


class FtolReached(TerminationCriterion):
    """Exception raised when the f_tol_rel or f_tol_abs criteria is reached by the
    driver."""


class XtolReached(TerminationCriterion):
    """Exception raised when the x_tol_rel or x_tol_abs criteria is reached by the
    driver."""


def is_x_tol_reached(opt_problem, x_tol_rel=1e-6, x_tol_abs=1e-6, n_x=2):
    """Tests if the tolerance on the design variables are reached The coordinate wise
    average of the last n_x points are taken Then it is checked that all points are
    within the distance of the center with relative and absolute tolerances specified by
    the user.

    Parameters
    ----------
    opt_problem: OptimizationProblem
        the optimization problem containing the iterations
    x_tol_rel: float
        relative tolerance
    x_tol_abs: float
        absolute tolerance
    n_x: int
        number of design vectors to account for
    """
    database = opt_problem.database
    if len(database) < n_x:
        return False
    x_list = database.get_last_n_x(n_x)

    # Checks that there is at least one feasible point
    is_feas = any((opt_problem.is_point_feasible(database[x]) for x in x_list))
    if not is_feas:
        return False
    x_average = average(x_list, axis=0)
    x_close = [allclose(x, x_average, atol=x_tol_abs, rtol=x_tol_rel) for x in x_list]
    return all(x_close)


def is_f_tol_reached(opt_problem, f_tol_rel=1e-6, f_tol_abs=1e-6, n_x=2):
    """Tests if the tolerance on the objective function are reached The average function
    value of the last n_x points are taken Then it is checked that all points are within
    the distance of the center with relative and absolute tolerances specified by the
    user.

    Parameters
    ----------
    opt_problem: OptimizationProblem
        the optimization problem containing the iterations
    x_tol_rel: float
        relative tolerance
    x_tol_abs: float
        absolute tolerance
    n_x: int
        number of design vectors to account for
    """
    database = opt_problem.database
    if len(database) < n_x:
        return False

    # Checks that there is at least one feasible point
    x_list = database.get_last_n_x(n_x)
    is_feas = any((opt_problem.is_point_feasible(database[x]) for x in x_list))
    if not is_feas:
        return False
    obj_name = opt_problem.objective.name
    f_list = [database.get_f_of_x(obj_name, x) for x in x_list]
    f_list = [f_val for f_val in f_list if f_val is not None]
    if not f_list:
        return False
    f_average = average(f_list)

    close = [allclose(f, f_average, atol=f_tol_abs, rtol=f_tol_rel) for f in f_list]

    return all(close)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Syver Doving Agdestein
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""The k-means algorithm for clustering.

The k-means algorithm groups the data into clusters,
where the number of clusters :math:`k` is fixed.
This is done by initializing :math:`k` centroids in the design space.
The points are grouped into clusters according to their nearest centroid.

When fitting the algorithm,
each centroid is successively moved to the mean of its corresponding cluster,
and the cluster value of each point is then reset
to the cluster value of the closest centroid.
This process is repeated until convergence.

Cluster values of new points may be predicted
by returning the value of the closest centroid.
Denoting :math:`(c_1, \\cdots, c_k) \\in \\mathbb{R}^{n \\times k}` the centroids,
and assuming no overlap between the centroids,
we may compute the prediction

.. math::

        \\operatorname{cluster}(x) =
            \\underset{i=1,\\cdots,k}{\\operatorname{argmin}} \\|x-c_i\\|.

A probability measure may also be provided,
using the distances from the point to each of the centroids:

.. math::

    \\mathbb{P}(x \\in C_i) = \\begin{cases}
        1 & \\operatorname{if} x = c_i\\\\
        0 & \\operatorname{if} x = c_j,\\ j \\neq i\\\\
        \\frac{\\frac{1}{\\|x-c_i\\|}}{\\sum_{j=1}^k \\frac{1}{\\|x-c_j\\|}}
            & \\operatorname{if} x \\neq c_j\\, \\forall j=1,\\cdots,k
    \\end{cases},

where :math:`C_i = \\{x\\, | \\, \\operatorname{cluster}(x) = i \\}`.
Here,
:math:`\\mathbb{P}(x \\in C_i)` represents the probability of cluster :math:`i`
given the point :math:`x`.


This concept is implemented through the :class:`.KMeans` class
which inherits from the :class:`.MLClusteringAlgo` class.

Dependence
----------
This clustering algorithm relies on the KMeans class
of the `scikit-learn library <https://scikit-learn.org/stable/modules/
generated/sklearn.cluster.KMeans.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Iterable, Optional, Union

from numpy import finfo, ndarray
from numpy.linalg import norm
from sklearn.cluster import KMeans as SKLKmeans

from gemseo.core.dataset import Dataset
from gemseo.mlearning.cluster.cluster import MLPredictiveClusteringAlgo
from gemseo.mlearning.core.ml_algo import TransformerType

LOGGER = logging.getLogger(__name__)


class KMeans(MLPredictiveClusteringAlgo):
    """The k-means clustering algorithm."""

    ABBR = "K-means"

    EPS = finfo(float).eps

    def __init__(
        self,
        data,  # type:Dataset
        transformer=None,  # type: Optional[TransformerType]
        var_names=None,  # type: Optional[Iterable[str]]
        n_clusters=5,  # type: int
        random_state=0,  # type: Optional[int]
        **parameters  # type: Optional[Union[int,float,bool,str]]
    ):  # type: (...) -> None
        """
        Args:
            n_clusters: The number of clusters of the K-means algorithm.
            random_state: If None, use a random generation of the initial centroids.
                If not None,
                the integer is used to make the initialization deterministic.
        """
        super(KMeans, self).__init__(
            data,
            transformer=transformer,
            var_names=var_names,
            n_clusters=n_clusters,
            random_state=random_state,
            **parameters
        )
        self.algo = SKLKmeans(n_clusters, random_state=random_state, **parameters)

    def _fit(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> None
        self.labels = self.algo.fit_predict(data)

    def _predict(
        self,
        data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return self.algo.predict(data)

    def _predict_proba_soft(
        self,
        data,  # type: ndarray
    ):  # type: (...)-> ndarray
        centers = self.algo.cluster_centers_
        distances = norm(data[:, None] - centers, axis=2)
        inverse_distances = 1 / (distances + self.EPS)
        probas = inverse_distances / inverse_distances.sum(axis=1)[:, None]
        return probas



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                         documentation
#        :author: Francois Gallard, Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
r"""The linear model algorithm for regression.

The linear regression surrogate discipline expresses the model output
as a weighted sum of the model inputs:

.. math::

    y = w_0 + w_1x_1 + w_2x_2 + ... + w_dx_d
    + \alpha \left( \lambda \|w\|_2 + (1-\lambda) \|w\|_1 \right),

where the coefficients :math:`(w_1, w_2, ..., w_d)` and the intercept
:math:`w_0` are estimated by least square regression. They are are easily
accessible via the arguments *coefficients* and *intercept*.

The penalty level :math:`\alpha` is a non-negative parameter intended to
prevent overfitting, while the penalty ratio :math:`\lambda\in [0, 1]`
expresses the ratio between :math:`\ell_2`- and :math:`\ell_1`-regularization.
When :math:`\lambda=1`, there is no :math:`\ell_1`-regularization, and a Ridge
regression is performed. When :math:`\lambda=0`, there is no
:math:`\ell_2`-regularization, and a Lasso regression is performed. For
:math:`\lambda` between 0 and 1, an elastic net regression is performed.

One may also choose not to penalize the regression at all, by setting
:math:`\alpha=0`. In this case, a simple least squares regression is performed.

This concept is implemented through the :class:`.LinearRegression` class which
inherits from the :class:`.MLRegressionAlgo` class.

Dependence
----------
The linear model relies on the LinearRegression, Ridge, Lasso and ElasticNet
classes of the `scikit-learn library <https://scikit-learn.org/stable/modules/
linear_model.html>`_.
"""
from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Optional, Union

from numpy import array, ndarray, repeat, zeros
from sklearn.linear_model import ElasticNet, Lasso
from sklearn.linear_model import LinearRegression as LinReg
from sklearn.linear_model import Ridge

from gemseo.core.dataset import Dataset
from gemseo.mlearning.core.ml_algo import DataType, TransformerType
from gemseo.mlearning.regression.regression import MLRegressionAlgo
from gemseo.mlearning.transform.dimension_reduction.dimension_reduction import (
    DimensionReduction,
)
from gemseo.utils.data_conversion import DataConversion

LOGGER = logging.getLogger(__name__)


class LinearRegression(MLRegressionAlgo):
    """Linear regression."""

    LIBRARY = "scikit-learn"
    ABBR = "LinReg"

    def __init__(
        self,
        data,  # type: Dataset
        transformer=None,  # type: Optional[TransformerType]
        input_names=None,  # type: Optional[Iterable[str]]
        output_names=None,  # type: Optional[Iterable[str]]
        fit_intercept=True,  # type: bool
        penalty_level=0.0,  # type: float
        l2_penalty_ratio=1.0,  # type: float
        **parameters  # type: Optional[Union[float,int,str,bool]]
    ):  # type: (...) ->None
        """
        Args:
            fit_intercept: If True, fit intercept.
            penalty_level: The penalty level greater or equal to 0.
                If 0, there is no penalty.
            l2_penalty_ratio: The penalty ratio related to the l2 regularization.
                If 1, use the Ridge penalty.
                If 0, use the Lasso penalty.
                Between 0 and 1, use the ElasticNet penalty.
            **parameters: The parameters of the machine learning algorithm.
        """
        super(LinearRegression, self).__init__(
            data,
            transformer=transformer,
            input_names=input_names,
            output_names=output_names,
            fit_intercept=fit_intercept,
            penalty_level=penalty_level,
            l2_penalty_ratio=l2_penalty_ratio,
            **parameters
        )
        if "degree" in parameters:
            del parameters["degree"]

        if penalty_level == 0.0:
            self.algo = LinReg(
                normalize=False, copy_X=False, fit_intercept=fit_intercept, **parameters
            )
        else:
            if l2_penalty_ratio == 1.0:
                self.algo = Ridge(
                    normalize=False,
                    copy_X=False,
                    fit_intercept=fit_intercept,
                    alpha=penalty_level,
                    **parameters
                )
            elif l2_penalty_ratio == 0.0:
                self.algo = Lasso(
                    normalize=False,
                    copy_X=False,
                    fit_intercept=fit_intercept,
                    alpha=penalty_level,
                    **parameters
                )
            else:
                self.algo = ElasticNet(
                    normalize=False,
                    copy_X=False,
                    fit_intercept=fit_intercept,
                    alpha=penalty_level,
                    l1_ratio=1 - l2_penalty_ratio,
                    **parameters
                )

    def _fit(
        self,
        input_data,  # type: ndarray
        output_data,  # type: ndarray
    ):  # type: (...) -> None
        self.algo.fit(input_data, output_data)

    def _predict(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        return self.algo.predict(input_data)

    def _predict_jacobian(
        self,
        input_data,  # type: ndarray
    ):  # type: (...) -> ndarray
        n_samples = input_data.shape[0]
        return repeat(self.algo.coef_[None], n_samples, axis=0)

    @property
    def coefficients(self):  # type: (...) ->ndarray
        """The regression coefficients of the linear model."""
        return self.algo.coef_

    @property
    def intercept(self):  # type: (...) ->ndarray
        """The regression intercepts of the linear model."""
        if self.parameters["fit_intercept"]:
            intercept = self.algo.intercept_
        else:
            intercept = zeros(self.algo.coef_.shape[0])
        return intercept

    def get_coefficients(
        self,
        as_dict=True,  # type: bool
    ):  # type: (...) -> DataType
        """Return the regression coefficients of the linear model.

        Args:
            as_dict: If True, return the coefficients as a dictionary.
                Otherwise, return the coefficients as a numpy.array

        Returns:
            The regression coefficients of the linear model.

        Raises:
            ValueError: If the coefficients are required as a dictionary
                even though the transformers change the variables dimensions.
        """
        coefficients = self.coefficients
        if as_dict:
            if any(
                [
                    isinstance(transformer, DimensionReduction)
                    for _, transformer in self.transformer.items()
                ]
            ):
                raise ValueError(
                    "Coefficients are only representable in dictionary "
                    "form if the transformers do not change the "
                    "dimensions of the variables."
                )
            coefficients = self.__convert_array_to_dict(coefficients)
        return coefficients

    def get_intercept(
        self,
        as_dict=True,  # type:bool
    ):  # type: (...) -> DataType
        """Return the regression intercepts of the linear model.

        Args:
            as_dict: If True, return the intercepts as a dictionary.
                Otherwise, return the intercepts as a numpy.array

        Returns:
            The regression intercepts of the linear model.

        Raises:
            ValueError: If the coefficients are required as a dictionary
                even though the transformers change the variables dimensions.
        """
        intercept = self.intercept
        if as_dict:
            if Dataset.OUTPUT_GROUP in self.transformer:
                raise ValueError(
                    "Intercept is only representable in dictionary "
                    "form if the transformers do not change the "
                    "dimensions of the output variables."
                )
            varsizes = self.learning_set.sizes
            intercept = DataConversion.array_to_dict(
                intercept, self.output_names, varsizes
            )
            intercept = {key: list(val) for key, val in intercept.items()}
        return intercept

    def __convert_array_to_dict(
        self,
        data,  # type:ndarray
    ):  # type: (...) -> Dict[str,ndarray]
        """Convert a data array into a dictionary.

        Args:
            data: The data to be converted.

        Returns:
            The converted data.
        """
        varsizes = self.learning_set.sizes
        data = [
            DataConversion.array_to_dict(row, self.input_names, varsizes)
            for row in data
        ]
        data = [{key: list(val) for key, val in element.items()} for element in data]
        data = DataConversion.array_to_dict(array(data), self.output_names, varsizes)
        data = {key: list(val) for key, val in data.items()}
        return data



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Dataset factory
===============

This module contains a factory
to instantiate a :class:`.Dataset` from its class name.
The class can be internal to |g| or located in an external module whose path
is provided to the constructor. It also provides a list of available cache
types and allows you to test if a cache type is available.
"""
from __future__ import division, unicode_literals

import logging

from gemseo.core.dataset import Dataset
from gemseo.core.factory import Factory

LOGGER = logging.getLogger(__name__)


class DatasetFactory(object):
    """This factory instantiates a :class:`.Dataset` from its class name.

    The class can be internal to |g| or located in an external module whose path is
    provided to the constructor.
    """

    def __init__(self):
        """Initializes the factory: scans the directories to search for subclasses of
        Dataset.

        Searches in "GEMSEO_PATH" and gemseo.mlearning.p_datasets
        """
        self.factory = Factory(Dataset, ("gemseo.problems.dataset",))

    def create(self, dataset, **options):
        """Create a dataset.

        :param str dataset: name of the dataset (its classname).
        :param options: additional options specific
        :return: dataset
        :rtype: Dataset
        """
        return self.factory.create(dataset, **options)

    @property
    def datasets(self):
        """Lists the available datasets.

        :returns: the list of datasets.
        :rtype: list(str)
        """
        return self.factory.classes

    def is_available(self, dataset):
        """Checks the availability of a dataset.

        :param str dataset:  name of the dataset (its class name).
        :returns: True if the dataset is available.
        :rtype: bool
        """
        return self.factory.is_available(dataset)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#                           documentation
#        :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Class for the empirical estimation of statistics from a dataset.

Overview
--------

The :class:`.EmpiricalStatistics` class inherits
from the abstract :class:`.Statistics` class
and aims to estimate statistics from a :class:`.Dataset`,
based on empirical estimators.

Construction
------------

A :class:`.EmpiricalStatistics` is built from a :class:`.Dataset`
and optionally variables names.
In this case,
statistics are only computed for these variables.
Otherwise,
statistics are computed for all the variable available in the dataset.
Lastly,
the user can give a name to its :class:`.EmpiricalStatistics` object.
By default,
this name is the concatenation of 'EmpiricalStatistics'
and the name of the :class:`.Dataset`.
"""

from __future__ import division, unicode_literals

import logging
from typing import Dict, Iterable, Optional

from numpy import all as np_all
from numpy import max as np_max
from numpy import mean
from numpy import min as np_min
from numpy import ndarray, quantile, std, var
from scipy.stats import moment

from gemseo.core.dataset import Dataset
from gemseo.uncertainty.statistics.statistics import Statistics

LOGGER = logging.getLogger(__name__)


class EmpiricalStatistics(Statistics):
    """Empirical estimation of statistics.

    Examples:
        >>> from gemseo.api import (
        ...     create_discipline,
        ...     create_parameter_space,
        ...     create_scenario)
        >>> from gemseo.uncertainty.statistics.empirical import EmpiricalStatistics
        >>>
        >>> expressions = {"y1": "x1+2*x2", "y2": "x1-3*x2"}
        >>> discipline = create_discipline(
        ...     "AnalyticDiscipline", expressions_dict=expressions
        ... )
        >>> discipline.set_cache_policy(discipline.MEMORY_FULL_CACHE)
        >>>
        >>> parameter_space = create_parameter_space()
        >>> parameter_space.add_random_variable(
        ...     "x1", "OTUniformDistribution", minimum=-1, maximum=1
        ... )
        >>> parameter_space.add_random_variable(
        ...     "x2", "OTUniformDistribution", minimum=-1, maximum=1
        ... )
        >>>
        >>> scenario = create_scenario(
        ...     [discipline],
        ...     "DisciplinaryOpt",
        ...     "y1",
        ...     parameter_space,
        ...     scenario_type="DOE"
        ... )
        >>> scenario.execute({'algo': 'OT_MONTE_CARLO', 'n_samples': 100})
        >>>
        >>> dataset = discipline.cache.export_to_dataset()
        >>>
        >>> statistics = EmpiricalStatistics(dataset)
        >>> mean = statistics.mean()
    """

    def __init__(
        self,
        dataset,  # type: Dataset,
        variables_names=None,  # type: Optional[Iterable[str]]
        name=None,  # type: Optional[str]
    ):  # type: (...) -> None # noqa: D107,D205,D212,D415
        name = name or dataset.name
        super(EmpiricalStatistics, self).__init__(dataset, variables_names, name)

    def compute_maximum(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {name: np_max(self.dataset[name], 0) for name in self.names}
        return result

    def compute_mean(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {name: mean(self.dataset[name], 0) for name in self.names}
        return result

    def compute_minimum(self):  # type: (...) -> Dict[str, ndarray] # noqa: D102
        result = {name: np_min(self.dataset[name], 0) for name in self.names}
        return result

    def compute_probability(
        self,
        thresh,  # type: float
        greater=True,  # type: bool
    ):  # type: (...) -> Dict[str,ndarray]  # noqa: D102
        if greater:
            result = {
                name: mean(np_all(self.dataset[name] >= thresh[name], 1))
                for name in self.names
            }
        else:
            result = {
                name: mean(np_all(self.dataset[name] <= thresh[name], 1))
                for name in self.names
            }
        return result

    def compute_quantile(
        self,
        prob,  # type:float
    ):  # type: (...) -> Dict[str, ndarray] # noqa: D102
        result = {name: quantile(self.dataset[name], prob, 0) for name in self.names}
        return result

    def compute_standard_deviation(
        self,
    ):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {name: std(self.dataset[name], 0) for name in self.names}
        return result

    def compute_variance(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {name: var(self.dataset[name], 0) for name in self.names}
        return result

    def compute_moment(
        self,
        order,  # type: int
    ):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        result = {name: moment(self.dataset[name], order, 0) for name in self.names}
        return result

    def compute_range(self):  # type: (...) -> Dict[str, ndarray]  # noqa: D102
        lower = self.compute_minimum()
        upper = self.compute_maximum()
        result = {name: upper[name] - lower[name] for name in self.names}
        return result



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or initial
#         documentation
#        :author:  Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
Scalable MDO problem
====================

This module implements the concept of scalable problem by means of the
:class:`.ScalableProblem` class.

Given

- a MDO scenario based on a set of sampled disciplines
  with a particular problem dimension,
- a new problem dimension (= number of inputs and outputs),

a scalable problem:

1. makes each discipline scalable based on the new problem dimension,
2. creates the corresponding MDO scenario.

Then, this MDO scenario can be executed and post-processed.

We can repeat this tasks for different sizes of variables
and compare the scalability, which is the dependence of the scenario results
on the problem dimension.

.. seealso:: :class:`.MDODiscipline`, :class:`.ScalableDiscipline`
   and :class:`.Scenario`
"""
from __future__ import division, unicode_literals

import logging
import os
from copy import deepcopy

from numpy import array, ones, random, where, zeros

from gemseo.algos.design_space import DesignSpace
from gemseo.api import (
    create_design_space,
    create_scenario,
    generate_coupling_graph,
    generate_n2_plot,
    get_all_inputs,
)
from gemseo.core.coupling_structure import MDOCouplingStructure
from gemseo.mda.mda_factory import MDAFactory
from gemseo.problems.scalable.data_driven.discipline import ScalableDiscipline
from gemseo.utils.string_tools import MultiLineString

LOGGER = logging.getLogger(__name__)


class ScalableProblem(object):
    """Scalable problem."""

    def __init__(
        self,
        datasets,
        design_variables,
        objective_function,
        eq_constraints=None,
        ineq_constraints=None,
        maximize_objective=False,
        sizes=None,
        **parameters
    ):
        """Constructor.

        :param list(Dataset) datasets: disciplinary datasets.
        :param list(str) design_variables: list of design variable names
        :param str objective_function: objective function
        :param list(str) eq_constraints: equality constraints. Default: None.
        :param list(str) eq_constraints: inequality constraints. Default: None.
        :param bool maximize_objective: maximize objective. Default: False.
        :param dict sizes: sizes of input and output variables.
            If None, use the original sizes. Default: None.
        :param parameters: optional parameters for the scalable model.
        """
        self.disciplines = [dataset.name for dataset in datasets]
        self.data = {dataset.name: dataset for dataset in datasets}
        self.inputs = {
            dataset.name: dataset.get_names(dataset.INPUT_GROUP) for dataset in datasets
        }
        self.outputs = {
            dataset.name: dataset.get_names(dataset.OUTPUT_GROUP)
            for dataset in datasets
        }
        self.varsizes = {}
        for dataset in datasets:
            self.varsizes.update(dataset.sizes)
        self.design_variables = design_variables
        self.objective_function = objective_function
        self.ineq_constraints = ineq_constraints
        self.eq_constraints = eq_constraints
        self.maximize_objective = maximize_objective
        self.scaled_disciplines = []
        self.scaled_sizes = {}
        self._build_scalable_disciplines(sizes, **parameters)
        self.scenario = None

    def __str__(self):
        """String representation of information about the scalable problem.

        :return: scalable problem description
        :rtype: str
        """
        disciplines = ", ".join(self.disciplines)
        design_variables = None
        if self.design_variables is not None:
            design_variables = ", ".join(self.design_variables)
        ineq_constraints = None
        if self.ineq_constraints is not None:
            ineq_constraints = ", ".join(self.ineq_constraints)
        eq_constraints = None
        if self.eq_constraints is not None:
            eq_constraints = ", ".join(self.eq_constraints)
        sizes = [
            name + " ({})".format(size) for name, size in self.scaled_sizes.items()
        ]
        sizes = ", ".join(sizes)
        optimize = "maximize" if self.maximize_objective else "minimize"
        msg = MultiLineString()
        msg.add("MDO problem")
        msg.indent()
        msg.add("Disciplines: {}", disciplines)
        msg.add("Design variables: {}", design_variables)
        msg.add("Objective function: {} (to {})", self.objective_function, optimize)
        msg.add("Inequality constraints: {}", ineq_constraints)
        msg.add("Equality constraints: {}", eq_constraints)
        msg.add("Sizes: {}", sizes)
        return str(msg)

    def plot_n2_chart(self, save=True, show=False):
        """Plot a N2 chart.

        :param bool save: save plot. Default: True.
        :param bool show: show plot. Default: False.
        """
        generate_n2_plot(self.scaled_disciplines, save=save, show=show)

    def plot_coupling_graph(self):
        """Plot a coupling graph."""
        generate_coupling_graph(self.scaled_disciplines)

    def plot_1d_interpolations(
        self, save=True, show=False, step=0.01, varnames=None, directory=".", png=False
    ):
        """Plot 1d interpolations.

        :param bool save: save plot. Default: True.
        :param bool show: show plot. Default: False.
        :param bool step: Step to evaluate the 1d interpolation function
            Default: 0.01.
        :param list(str) varnames: names of the variable to plot;
            if None, all variables are plotted. Default: None.
        :param str directory: directory path. Default: '.'.
        :param bool png: if True, the file format is PNG. Otherwise, use PDF.
            Default: False.
        """
        if not os.path.exists(directory):
            os.mkdir(directory)
        allfnames = []
        for scalable_discipline in self.scaled_disciplines:
            func = scalable_discipline.scalable_model.plot_1d_interpolations
            fnames = func(save, show, step, varnames, directory, png)
            allfnames = allfnames + [os.path.join(directory, fname) for fname in fnames]
        return allfnames

    def plot_dependencies(self, save=True, show=False, directory="."):
        """Plot dependency matrices.

        :param bool save: save plot (default: True)
        :param bool show: show plot (default: False)
        :param str directory: directory path (default: '.')
        """
        fnames = []
        for scalable_discipline in self.scaled_disciplines:
            scalable_model = scalable_discipline.scalable_model
            plot_dependency = scalable_model.plot_dependency
            fname = plot_dependency(
                add_levels=True, save=save, show=show, directory=directory
            )
            fnames.append(fname)
        return fnames

    def _build_scalable_disciplines(self, sizes=None, **parameters):
        """Build scalable disciplines.

        :param dict sizes: dictionary whose keys are variable names
            and variables sizes.
        :param parameters: options.
        """
        copied_parameters = deepcopy(parameters)
        for disc in self.disciplines:
            varnames = self.inputs[disc] + self.outputs[disc]
            sizes = sizes or {}
            new_varsizes = {
                varname: sizes.get(varname, self.varsizes[varname])
                for varname in varnames
            }
            if "group_dep" in parameters:
                copied_parameters["group_dep"] = parameters["group_dep"][disc]
            if "fill_factor" in parameters:
                copied_parameters["fill_factor"] = parameters["fill_factor"][disc]
            self.scaled_disciplines.append(
                ScalableDiscipline(
                    "ScalableDiagonalModel",
                    self.data[disc],
                    new_varsizes,
                    **copied_parameters
                )
            )
            self.scaled_sizes.update(deepcopy(new_varsizes))

    def create_scenario(
        self,
        formulation="DisciplinaryOpt",
        scenario_type="MDO",
        start_at_equilibrium=False,
        active_probability=0.1,
        feasibility_level=0.5,
        **options
    ):
        """Create MDO scenario from the scalable disciplines.

        :param str formulation: MDO formulation. Default: 'DisciplinaryOpt'.
        :param str scenario_type: type of scenario ('MDO' or 'DOE').
            Default: 'MDO'.
        :param bool start_at_equilibrium: start at equilibrium
            using a preliminary MDA. Default: True.
        :param float active_probability: probability to set the inequality
            constraints as active at initial step of the optimization.
            Default: 0.1.
        :param float feasibility_level: offset of satisfaction for inequality
            constraints. Default: 0.5.
        :param options: formulation options.
        """
        equilibrium = None
        if start_at_equilibrium:
            equilibrium = self.__get_equilibrium()
        disciplines = self.scaled_disciplines
        design_space = self._create_design_space(disciplines, formulation)
        max_obj = self.maximize_objective
        if formulation == "BiLevel":
            self.scenario = self._create_bilevel_scenario(disciplines, **options)
        else:
            self.scenario = create_scenario(
                disciplines,
                formulation,
                self.objective_function,
                deepcopy(design_space),
                scenario_type=scenario_type,
                maximize_objective=max_obj,
                **options
            )
        equilibrium = {} if not isinstance(equilibrium, dict) else equilibrium
        self.__add_ineq_constraints(active_probability, feasibility_level, equilibrium)
        self.__add_eq_constraints(equilibrium)
        return self.scenario

    def _create_bilevel_scenario(self, disciplines, **sub_scenario_options):
        """Create a bilevel scenario from disciplines.

        :param list(MDODiscipline) disciplines: list of MDODiscipline
        """
        cpl_structure = MDOCouplingStructure(disciplines)
        st_cpl_disciplines = cpl_structure.strongly_coupled_disciplines()
        wk_cpl_disciplines = cpl_structure.weakly_coupled_disciplines()
        obj = self.objective_function
        max_obj = self.maximize_objective

        # Construction of the subsystem scenarios
        sub_scenarios = []
        sub_inputs = []
        for discipline in st_cpl_disciplines:
            cplt_disciplines = list(set(disciplines) - {discipline})
            sub_disciplines = [discipline] + wk_cpl_disciplines
            design_space = DesignSpace()
            inputs = get_all_inputs([discipline])
            all_inputs = get_all_inputs(cplt_disciplines)
            inputs = list(set(inputs) - set(all_inputs))
            sub_inputs += inputs
            for name in inputs:
                design_space.add_variable(
                    name, self.scaled_sizes[name], "float", 0.0, 1.0, 0.5
                )
            sub_scenarios.append(
                create_scenario(
                    disciplines=sub_disciplines,
                    formulation="DisciplinaryOpt",
                    objective_name=obj,
                    design_space=design_space,
                    maximize_objective=max_obj,
                )
            )
            sub_scenarios[-1].default_inputs = sub_scenario_options

        # Construction of the system scenario
        all_inputs = get_all_inputs(disciplines)
        inputs = list(set(all_inputs) - set(sub_inputs))
        design_space = DesignSpace()
        for name in inputs:
            design_space.add_variable(
                name, self.scaled_sizes[name], "float", 0.0, 1.0, 0.5
            )
        sub_disciplines = sub_scenarios + wk_cpl_disciplines
        system_scenario = create_scenario(
            disciplines=sub_disciplines,
            formulation="BiLevel",
            objective_name=obj,
            design_space=design_space,
            maximize_objective=max_obj,
            mda_name="MDAJacobi",
            tolerance=1e-8,
        )
        return system_scenario

    def _create_design_space(self, disciplines=None, formulation="DisciplinaryOpt"):
        """Create a design space into the unit hypercube.

        :param list(MDODiscipline) disciplines: list of MDODiscipline
        :param str formulation: MDO formulation (default: 'DisciplinaryOpt')
        """
        design_space = create_design_space()
        for varname in self.design_variables:
            size = self.scaled_sizes[varname]
            l_b = zeros(size)
            u_b = ones(size)
            value = 0.5 + zeros(size)
            design_space.add_variable(varname, size, "float", l_b, u_b, value)
        if formulation == "IDF":
            coupling_structure = MDOCouplingStructure(disciplines)
            all_couplings = set(coupling_structure.get_all_couplings())
            for varname in all_couplings:
                size = self.scaled_sizes[varname]
                design_space.add_variable(
                    varname, size, "float", zeros(size), ones(size), 0.5 + zeros(size)
                )
        return design_space

    def __get_equilibrium(self, mda_name="MDAJacobi", **options):
        """Get the equilibrium point from a MDA method.

        :param str mda_name: MDA name (default: 'MDAJacobi')
        :return: equilibrium point
        :rtype: dict
        """
        LOGGER.info("Build a preliminary MDA to start at equilibrium")
        factory = MDAFactory()
        mda = factory.create(mda_name, self.scaled_disciplines, **options)
        if len(mda.strong_couplings) == 0:
            mda = factory.create("MDAQuasiNewton", self.scaled_disciplines, **options)
        return mda.execute()

    def __add_ineq_constraints(
        self, active_probability, feasibility_level, equilibrium
    ):
        """Add inequality constraints.

        :param float active_probability: probability to set the inequality
            constraints as active at initial step of the optimization
        :param float feasibility_level: offset of satisfaction for inequality
            constraints
        :param dict equilibrium: starting point at equilibrium
        """
        if not hasattr(feasibility_level, "__len__"):
            feasibility_level = {
                constraint: feasibility_level for constraint in self.ineq_constraints
            }
        for constraint, alphai in feasibility_level.items():
            if constraint in list(equilibrium.keys()):
                sample = random.rand(len(equilibrium[constraint]))
                val = equilibrium[constraint]
                taui = where(
                    sample < active_probability, val, alphai + (1 - alphai) * val
                )
            else:
                taui = 0.0
            self.scenario.add_constraint(constraint, "ineq", value=taui)

    def __add_eq_constraints(self, equilibrium):
        """Add equality constraints.

        :param dict equilibrium: starting point at equilibrium
        """
        for constraint in self.eq_constraints:
            cstr_value = equilibrium.get(constraint, array([0.0]))[0]
            self.scenario.add_constraint(constraint, "eq", value=cstr_value)

    def exec_time(self, do_sum=True):
        """Get total execution time per discipline.

        :param bool do_sum: sum over disciplines (default: True)
        :return: execution time
        :rtype: list(float) or float
        """
        exec_time = [discipline.exec_time for discipline in self.scenario.disciplines]
        if do_sum:
            exec_time = sum(exec_time)
        return exec_time

    @property
    def n_calls_top_level(self):
        """Get number of top level disciplinary calls per discipline.

        :return: number of top level disciplinary calls per discipline
        :rtype: list(int) or int
        """
        disciplines = self.scenario.formulation.get_top_level_disc()
        n_calls = {discipline.name: discipline.n_calls for discipline in disciplines}
        return n_calls

    @property
    def n_calls_linearize_top_level(self):
        """Get number of top level disciplinary calls per discipline.

        :return: number of top level disciplinary calls per discipline
        :rtype: list(int) or int
        """
        disciplines = self.scenario.formulation.get_top_level_disc()
        n_calls = {
            discipline.name: discipline.n_calls_linearize for discipline in disciplines
        }
        return n_calls

    @property
    def n_calls(self):
        """Get number of disciplinary calls per discipline.

        :return: number of disciplinary calls per discipline
        :rtype: list(int) or int
        """
        n_calls = {
            discipline.name: discipline.n_calls
            for discipline in self.scenario.disciplines
        }
        return n_calls

    @property
    def n_calls_linearize(self):
        """Get number of disciplinary calls per discipline.

        :return: number of disciplinary calls per discipline
        :rtype: list(int) or int
        """
        tmp = {
            discipline.name: discipline.n_calls_linearize
            for discipline in self.scenario.disciplines
        }
        return tmp

    @property
    def status(self):
        """Get the status of the scenario."""
        return self.scenario.optimization_result.status

    @property
    def is_feasible(self):
        """Get the feasibility property of the scenario."""
        return self.scenario.optimization_result.is_feasible



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#       :author: Matthias De Lozzo
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Abstract class for the estimation of statistics from a dataset.

Overview
--------

The abstract :class:`.Statistics` class implements the concept of statistics library.
It is enriched by the :class:`.EmpiricalStatistics` and :class:`.ParametricStatistics`.

Construction
------------

A :class:`.Statistics` object is built from a :class:`.Dataset`
and optionally variables names.
In this case,
statistics are only computed for these variables.
Otherwise,
statistics are computed for all the variable available in the dataset.
Lastly,
the user can give a name to its :class:`.Statistics` object.
By default,
this name is the concatenation of the name of the class overloading :class:`.Statistics`
and the name of the :class:`.Dataset`.

Capabilities
------------

A :class:`.Statistics` returns standard descriptive and statistical measures
for the different variables:

- :meth:`.Statistics.compute_minimum`: the minimum value,
- :meth:`.Statistics.compute_maximum`: the maximum value,
- :meth:`.Statistics.compute_range`: the difference between minimum and maximum values,
- :meth:`.Statistics.compute_mean`: the expectation (a.k.a. mean value),
- :meth:`.Statistics.compute_moment`: a central moment,
  which is a the expected value
  of a specified integer power
  of the deviation from the mean,
- :meth:`.Statistics.compute_variance`: the variance,
  which is the mean squared variation around the mean value,
- :meth:`.Statistics.compute_standard_deviation`: the standard deviation,
  which is the square root of the variance,
- :meth:`.Statistics.compute_quantile`: the quantile associated with a probability,
  which is the cut point diving the range into a first continuous interval
  with this given probability and a second continuous interval
  with the complementary probability; common *q*-quantiles dividing
  the range into *q* continuous interval with equal probabilities are also implemented:

    - :meth:`.Statistics.compute_median`
      which implements the 2-quantile (50%).
    - :meth:`.Statistics.compute_quartile`
      whose order (1, 2 or 3) implements the 4-quantiles (25%, 50% and 75%),
    - :meth:`.Statistics.compute_percentile`
      whose order (1, 2, ..., 99) implements the 100-quantiles (1%, 2%, ..., 99%),

- :meth:`.Statistics.compute_probability`:
  the probability that the random variable is larger or smaller
  than a certain threshold,
- :meth:`.Statistics.compute_tolerance_interval`:
  the left-sided, right-sided or both-sided tolerance interval
  associated with a given coverage level and a given confidence level,
  which is a statistical interval within which,
  with some confidence level,
  a specified proportion of the random variable realizations falls
  (this proportion is the coverage level)

    - :meth:`.Statistics.compute_a_value`:
      the A-value, which is the lower bound of the left-sided tolerance interval
      associated with a coverage level equal to 99% and a confidence level equal to 95%,
    - :meth:`.Statistics.compute_b_value`:
      the B-value, which is the lower bound of the left-sided tolerance interval
      associated with a coverage level equal to 90% and a confidence level equal to 95%,
"""

from __future__ import division, unicode_literals

import logging
from enum import Enum
from typing import Dict, Iterable, Optional, Tuple

import six
from custom_inherit import DocInheritMeta
from numpy import ndarray

from gemseo.core.dataset import Dataset
from gemseo.uncertainty.statistics.tolerance_interval.distribution import (
    ToleranceIntervalSide,
)
from gemseo.utils.string_tools import MultiLineString, pretty_repr

LOGGER = logging.getLogger(__name__)


@six.add_metaclass(
    DocInheritMeta(
        abstract_base_class=True,
        style="google_with_merge",
        include_special_methods=True,
    )
)
class Statistics(object):
    """Abstract class to interface a statistics library.

    Attributes:
        dataset (Dataset): The dataset.
        n_samples (int): The number of samples.
        n_variables (int): The number of variables.
        name (str): The name of the object.
    """

    SYMBOLS = {}

    def __init__(
        self,
        dataset,  # type: Dataset,
        variables_names=None,  # type: Optional[Iterable[str]]
        name=None,  # type: Optional[str]
    ):  # type: (...) -> None # noqa: D205,D212,D415
        """
        Args:
            dataset: A dataset.
            variables_names: The variables of interest.
                Default: consider all the variables available in the dataset.
            name: A name for the object.
                Default: use the concatenation of the class and dataset names.
        """
        class_name = self.__class__.__name__
        default_name = "{}_{}".format(class_name, dataset.name)
        self.name = name or default_name
        msg = "Create {}, a {} library.".format(self.name, class_name)
        LOGGER.info(msg)
        self.dataset = dataset.get_all_data(by_group=False, as_dict=True)
        self.n_samples = dataset.n_samples
        self.names = variables_names or dataset.variables
        self.n_variables = dataset.n_variables

    def __str__(self):  # type: (...) -> str
        msg = MultiLineString()
        msg.add(self.name)
        msg.indent()
        msg.add("n_samples: {}", self.n_samples)
        msg.add("n_variables: {}", self.n_variables)
        msg.add("variables: {}", pretty_repr(self.names))
        return str(msg)

    def compute_tolerance_interval(
        self,
        coverage,  # type: float
        confidence=0.95,  # type: float
        side=ToleranceIntervalSide.BOTH,  # type: ToleranceIntervalSide
    ):  # type: (...) -> Dict[str, Tuple[ndarray,ndarray]]# noqa: D102
        r"""Compute a tolerance interval (TI) for a given coverage level.

        This coverage level is the minimum percentage of belonging to the TI.
        The tolerance interval is computed with a confidence level
        and can be either lower-sided, upper-sided or both-sided.

        Args:
            coverage: A minimum percentage of belonging to the TI.
            confidence: A level of confidence in [0,1].
            side: The type of the tolerance interval
                characterized by its *sides* of interest,
                either a lower-sided tolerance interval :math:`[a, +\infty[`,
                an upper-sided tolerance interval :math:`]-\infty, b]`,
                or a two-sided tolerance interval :math:`[c, d]`.

        Returns:
            The tolerance limits of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["tolerance_interval"] = "TI"

    def compute_a_value(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the A-value.

        Returns:
            The A-value of the different variables.
        """
        result = self.compute_tolerance_interval(
            1 - 0.1, 0.99, ToleranceIntervalSide.LOWER
        )
        result = {name: value[0] for name, value in result.items()}
        return result

    SYMBOLS["a_value"] = "Aval"

    def compute_b_value(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the B-value.

        Returns:
            The B-value of the different variables.
        """
        result = self.compute_tolerance_interval(
            1 - 0.1, 0.95, ToleranceIntervalSide.LOWER
        )
        result = {name: value[0] for name, value in result.items()}
        return result

    SYMBOLS["b_value"] = "Bval"

    def compute_maximum(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the maximum.

        Returns:
            The maximum of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["maximum"] = "Max"

    def compute_mean(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the mean.

        Returns:
            The mean of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["mean"] = "E"

    def compute_mean_std(
        self,
        std_factor,  # type: float
    ):  # type: (...) -> Dict[str,ndarray]
        """Compute mean + std_factor * std.

        Returns:
            mean + std_factor * std for the different variables.
        """
        result = self.compute_mean()
        for name, value in self.compute_standard_deviation().items():
            result[name] += std_factor * value
        return result

    SYMBOLS["mean_std"] = "E_StD"

    def compute_minimum(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the minimum.

        Returns:
            The minimum of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["minimum"] = "Min"

    def compute_median(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the median.

        Returns:
            The median of the different variables.
        """
        result = self.compute_quantile(0.5)
        return result

    SYMBOLS["median"] = "Med"

    def compute_percentile(
        self,
        order,  # type: int
    ):  # type: (...) -> Dict[str,ndarray]
        """Compute the n-th percentile.

        Args:
            order: The order of the percentile.
                Either 0, 1, 2, ... or 100.

        Returns:
            The percentile of the different variables.
        """
        if not isinstance(order, int) or order > 100 or order < 0:
            raise TypeError(
                "Percentile order must be an integer between 0 and 100 inclusive."
            )
        prob = order / 100.0
        result = self.compute_quantile(prob)
        return result

    SYMBOLS["percentile"] = "p"

    def compute_probability(
        self,
        thresh,  # type: float
        greater=True,  # type: bool
    ):  # type: (...) -> Dict[str,ndarray]
        """Compute the probability related to a threshold.

        Args:
            thresh: A threshold.
            greater: The type of probability.
                If True,
                compute the probability of exceeding the threshold.
                Otherwise,
                compute the opposite.

        Returns:
            The probability of the different variables
        """
        raise NotImplementedError

    SYMBOLS["probability"] = "P"

    def compute_quantile(
        self,
        prob,  # type:float
    ):  # type: (...) -> Dict[str,ndarray]
        """Compute the quantile related to a probability.

        Args:
            prob: A probability between 0 and 1.

        Returns:
            The quantile of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["quantile"] = "Q"

    def compute_quartile(
        self,
        order,  # type:int
    ):  # type: (...) -> Dict[str,ndarray]
        """Compute the n-th quartile.

        Args:
            order: The order of the quartile. Either 1, 2 or 3.

        Returns:
            The quartile of the different variables.
        """
        quartiles = [0.25, 0.5, 0.75]
        if order not in [1, 2, 3]:
            raise ValueError("Quartile order must be in [1,2,3]")
        prob = quartiles[order - 1]
        result = self.compute_quantile(prob)
        return result

    SYMBOLS["quartile"] = "q"

    def compute_range(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the range.

        Returns:
            The range of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["range"] = "R"

    def compute_standard_deviation(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the standard deviation.

        Returns:
            The standard deviation of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["standard_deviation"] = "StD"

    def compute_variance(self):  # type: (...) -> Dict[str,ndarray]
        """Compute the variance.

        Returns:
            The variance of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["variance"] = "V"

    def compute_moment(
        self,
        order,  # type:int
    ):  # type: (...) -> Dict[str,ndarray]
        """Compute the n-th moment.

        Args:
            order: The order of a moment.

        Returns:
            The moment of the different variables.
        """
        raise NotImplementedError

    SYMBOLS["moment"] = "M"

    @classmethod
    def compute_expression(
        cls,
        variable,  # type:str
        function,  # type:str
        show_name=False,  # type:bool
        **options
    ):  # type: (...) -> str
        """Return the expression of a statistical function applied to a variable.

        Args:
            variable: The name of the variable.
            function: The name of the function.
            show_name: If True, show name. Otherwise, only show value.
            **options: The options passed to the statistical function.

        Returns:
            The expression of the statistical function applied to the variable.
        """
        middle = ""
        if "greater" in options:
            middle = ">=" if options["greater"] else "<="
            options.pop("greater")
        elif function == "probability":
            middle = ">="
        if show_name:
            value = []
            for name in sorted(options):
                value.append("{}={}".format(name, options[name]))
        else:
            value = []
            for name in sorted(options):
                if isinstance(options[name], Enum):
                    value.append(str(options[name].name))
                else:
                    value.append(str(options[name]))
        value = ", ".join(value)
        if value != "" and middle == "":
            middle = "; "
        return "{}[{}{}{}]".format(cls.SYMBOLS[function], variable, middle, value)



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
from __future__ import division, unicode_literals

import logging
from copy import deepcopy
from multiprocessing import cpu_count

from numpy import atleast_2d, concatenate, dot
from numpy.linalg import lstsq

from gemseo.core.discipline import MDODiscipline
from gemseo.core.execution_sequence import ExecutionSequenceFactory
from gemseo.core.parallel_execution import DiscParallelExecution
from gemseo.mda.mda import MDA
from gemseo.utils.data_conversion import DataConversion

# Contributors:
#    INITIAL AUTHORS - API and implementation and/or documentation
#        :author: Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES
"""
A Jacobi algorithm for solving MDAs
***********************************
"""


LOGGER = logging.getLogger(__name__)
N_CPUS = cpu_count()


class MDAJacobi(MDA):
    """Perform a MDA analysis using a Jacobi algorithm, an iterative technique to solve
    the linear system:

    .. math::

       Ax = b

    by decomposing the matrix :math:`A`
    into the sum of a diagonal matrix :math:`D`
    and the reminder :math:`R`.

    The new iterate is given by:

    .. math::

       x_{k+1} = D^{-1}(b-Rx_k)
    """

    SECANT_ACCELERATION = "secant"
    M2D_ACCELERATION = "m2d"

    def __init__(
        self,
        disciplines,
        max_mda_iter=10,
        name=None,
        n_processes=N_CPUS,
        acceleration=M2D_ACCELERATION,
        tolerance=1e-6,
        linear_solver_tolerance=1e-12,
        use_threading=True,
        warm_start=False,
        use_lu_fact=False,
        grammar_type=MDODiscipline.JSON_GRAMMAR_TYPE,
    ):
        """Constructor.

        :param disciplines: the disciplines list
        :type disciplines: list(MDODiscipline)
        :param max_mda_iter: maximum number of iterations
        :type max_mda_iter: int
        :param name: the name of the chain
        :type name: str
        :param n_processes: maximum number of processors on which to run
        :type n_processes: int
        :param acceleration: type of acceleration to be used to extrapolate
            the residuals and save CPU time by reusing the information
            from the last iterations, either None, or m2d, or secant,
            m2d is faster but uses the 2 last iterations
        :type acceleration: str
        :param tolerance: tolerance of the iterative direct coupling solver,
            norm of the current residuals divided by initial residuals norm
            shall be lower than the tolerance to stop iterating
        :type tolerance: float
        :param linear_solver_tolerance: Tolerance of the linear solver
            in the adjoint equation
        :type linear_solver_tolerance: float
        :param use_threading: use multithreading for parallel executions
            otherwise use multiprocessing
        :type use_threading: bool
        :param warm_start: if True, the second iteration and ongoing
            start from the previous coupling solution
        :type warm_start: bool
        :param use_lu_fact: if True, when using adjoint/forward
            differenciation, store a LU factorization of the matrix
            to solve faster multiple RHS problem
        :type use_lu_fact: bool
        :param grammar_type: the type of grammar to use for IO declaration
            either JSON_GRAMMAR_TYPE or SIMPLE_GRAMMAR_TYPE
        :type grammar_type: str
        """
        self.n_processes = n_processes
        super(MDAJacobi, self).__init__(
            disciplines,
            max_mda_iter=max_mda_iter,
            name=name,
            tolerance=tolerance,
            linear_solver_tolerance=linear_solver_tolerance,
            warm_start=warm_start,
            use_lu_fact=use_lu_fact,
            grammar_type=grammar_type,
        )
        self._initialize_grammars()
        self._set_default_inputs()
        self._compute_input_couplings()
        self.acceleration = acceleration
        self._dx_n = []
        self._g_x_n = []
        self.sizes = None
        self.parallel_execution = DiscParallelExecution(
            disciplines, n_processes, use_threading
        )

    def _compute_input_couplings(self):
        """Compute all the coupling variables that are inputs of the MDA.

        This must be overloaded here because the Jacobi algorithm induces a delay
        between the couplings, the strong couplings may be fully resolved but the weak
        ones may need one more iteration. The base MDA class uses strong couplings only
        which is not satisfying here if all disciplines are not strongly coupled
        """
        if len(self.coupling_structure.strongly_coupled_disciplines()) == len(
            self.disciplines
        ):
            return super(MDAJacobi, self)._compute_input_couplings()

        inputs = self.get_input_data_names()
        strong_cpl = self.coupling_structure.get_all_couplings()
        self._input_couplings = set(strong_cpl) & set(inputs)

    def _initialize_grammars(self):
        """Defines all inputs and outputs of the chain."""
        for discipline in self.disciplines:
            self.input_grammar.update_from(discipline.input_grammar)
            self.output_grammar.update_from(discipline.output_grammar)

    def execute_all_disciplines(self, input_local_data):
        """Executes all self.disciplines.

        :param input_local_data: the input data of the disciplines
        """
        self.reset_disciplines_statuses()

        if self.n_processes > 1:
            n_disc = len(self.disciplines)
            inputs_copy_list = [deepcopy(input_local_data) for _ in range(n_disc)]
            self.parallel_execution.execute(inputs_copy_list)
        else:
            for disc in self.disciplines:
                disc.execute(deepcopy(input_local_data))

        outputs = [discipline.get_output_data() for discipline in self.disciplines]
        for data in outputs:
            self.local_data.update(data)

    def get_expected_workflow(self):
        """See MDA.get_expected_workflow."""
        sub_workflow = ExecutionSequenceFactory.serial(self.disciplines)
        if self.n_processes > 1:
            sub_workflow = ExecutionSequenceFactory.parallel(self.disciplines)
        return ExecutionSequenceFactory.loop(self, sub_workflow)

    def _run(self):
        """Run method of the chain: executes all disciplines in a loop until outputs
        converge. Stops when.

        ||outputs-previous output||/||first outputs|| < self.tolerance

        :returns: the local data updated
        """
        if self.warm_start:
            self._couplings_warm_start()
        self._dx_n = []
        self._g_x_n = []
        # execute the disciplines
        current_couplings = self._current_input_couplings()
        self.execute_all_disciplines(deepcopy(self.local_data))
        new_couplings = self._current_input_couplings()
        self._dx_n.append(new_couplings - current_couplings)
        self._g_x_n.append(new_couplings)

        # store initial residual
        current_iter = 1
        self._compute_residual(
            current_couplings, new_couplings, current_iter, first=True
        )
        current_couplings = new_couplings

        while not self._termination(current_iter):
            self.execute_all_disciplines(deepcopy(self.local_data))
            new_couplings = self._current_input_couplings()

            # store current residual
            current_iter += 1
            self._compute_residual(current_couplings, new_couplings, current_iter)
            x_np1 = self._compute_nex_iterate(current_couplings, new_couplings)
            current_couplings = x_np1

    def _compute_nex_iterate(self, current_couplings, new_couplings):
        """Compute the next iterate given the evaluation of the couplings Eventually
        computes the secant method acceleration, see.

        See :
        Iterative residual-based vector methods to accelerate
        fixed point iterations, Isabelle Ramiere, Thomas Helfer

        :param current_couplings: input couplings of the disciplines
            given for evaluation at the last iterations
        :param current_couplings: computed couplings of the disciplines
            at the last iterations
        """

        self._dx_n.append(new_couplings - current_couplings)
        self._g_x_n.append(new_couplings)
        coupl_names = self._input_couplings
        if self.sizes is None:
            self.sizes = {key: self.local_data[key].size for key in coupl_names}
        dxn = self._dx_n[-1]
        dxn_1 = self._dx_n[-2]
        g_n = self._g_x_n[-1]
        gn_1 = self._g_x_n[-2]
        x_np1 = new_couplings
        if self.acceleration == self.SECANT_ACCELERATION:
            x_np1 = self._compute_secant_acc(dxn, dxn_1, g_n, gn_1)
        elif self.acceleration == self.M2D_ACCELERATION:
            if len(self._dx_n) >= 3:
                dxn_2 = self._dx_n[-3]
                dgn_2 = self._g_x_n[-3]
            else:
                dxn_2 = self._dx_n[-2]
                dgn_2 = self._g_x_n[-2]

            x_np1 = self._compute_m2d_acc(dxn, dxn_1, dxn_2, g_n, gn_1, dgn_2)

        if len(self._dx_n) > 3:  # Forget too old stuff
            self._dx_n = self._dx_n[-3:]
            self._g_x_n = self._g_x_n[-3:]
        new_c = DataConversion.array_to_dict(x_np1, coupl_names, self.sizes)
        self.local_data.update(new_c)
        return x_np1

    @staticmethod
    def _minimize_2md(dxn, dxn_1, dxn_2):
        """Compute the extrapolation coefficients of the 2-delta method Minimizes the
        sub problem in the d-2 method Use a least squares solver to find he minimizer
        of.

        dxn - x[0] * (dxn - dxn_1) - x[1] * (dxn_1 - dxn_2)

        :param dxn: delta couplings at last iteration
        :param dxn_1: delta couplings at last iteration-1
        :param dxn_2: delta couplings at last iteration-2
        :returns: lambda
        """
        mat = concatenate((atleast_2d(dxn - dxn_1), atleast_2d(dxn_1 - dxn_2)))
        return lstsq(mat.T, dxn, rcond=None)[0]

    @staticmethod
    def _compute_secant_acc(dxn, dxn_1, cgn, cgn_1):
        """secant acceleration.

        from the paper:
        "Iterative residual-based vector methods to accelerate
        fixed point iterations",  Isabelle Ramiere, Thomas Helfer

        secant acceleration: page 15 equation (41)

        :param dxn: delta couplings at last iteration
        :param dxn_1: delta couplings at last iteration-1
        :param cgn: computed couplings at last iteration
        :param cgn_1: computed couplings at last iteration-1
        """
        d_dxn = dxn - dxn_1
        acc = (cgn - cgn_1) * dot(d_dxn, dxn) / dot(d_dxn, d_dxn)
        return cgn - acc

    def _compute_m2d_acc(self, dxn, dxn_1, dxn_2, g_n, gn_1, gn_2):
        """2-delta acceleration.

        from the paper:
        "Iterative residual-based vector methods to accelerate
        fixed point iterations",  Isabelle Ramiere, Thomas Helfer
        page 22 eq (50)

        :param dxn: delta couplings at last iteration
        :param dxn_1: delta couplings at last iteration-1
        :param dxn_2: delta couplings at last iteration-2
        :param g_n: computed couplings at last iteration
        :param gn_1: computed couplings at last iteration-1
        :param gn_2: computed couplings at last iteration-2
        """
        lamba_min = self._minimize_2md(dxn, dxn_1, dxn_2)
        acc = lamba_min[0] * (g_n - gn_1) + lamba_min[1] * (gn_1 - gn_2)
        return g_n - acc



# -*- coding: utf-8 -*-
from __future__ import  unicode_literals

from .draft06 import CodeGeneratorDraft06


class CodeGeneratorDraft07(CodeGeneratorDraft06):
    FORMAT_REGEXS = dict(CodeGeneratorDraft06.FORMAT_REGEXS, **{
        'date': r'^(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})\Z',
        'iri': r'^\w+:(\/?\/?)[^\s]+\Z',
        'iri-reference': r'^(\w+:(\/?\/?))?[^#\\\s]*(#[^\\\s]*)?\Z',
        'idn-email': r'^[^@]+@[^@]+\.[^@]+\Z',
        #'idn-hostname': r'',
        'relative-json-pointer': r'^(?:0|[1-9][0-9]*)(?:#|(?:\/(?:[^~/]|~0|~1)*)*)\Z',
        #'regex': r'',
        'time': (
            r'^(?P<hour>\d{1,2}):(?P<minute>\d{1,2})'
            r'(?::(?P<second>\d{1,2})(?:\.(?P<microsecond>\d{1,6}))?'
            r'([zZ]|[+-]\d\d:\d\d)?)?\Z'
        ),
    })

    def __init__(self, definition, resolver=None):
        super(CodeGeneratorDraft07, self).__init__(definition, resolver)
        # pylint: disable=duplicate-code
        self._json_keywords_to_function.update((
            ('if', self.generate_if_then_else),
            ('contentEncoding', self.generate_content_encoding),
            ('contentMediaType', self.generate_content_media_type),
        ))

    def generate_if_then_else(self):
        """
        Implementation of if-then-else.

        .. code-block:: python

            {
                'if': {
                    'exclusiveMaximum': 0,
                },
                'then': {
                    'minimum': -10,
                },
                'else': {
                    'multipleOf': 2,
                },
            }

        Valid values are any between -10 and 0 or any multiplication of two.
        """
        with self.l('try:'):
            self.generate_func_code_block(
                self._definition['if'],
                self._variable,
                self._variable_name,
                clear_variables=True
            )
        with self.l('except JsonSchemaException:'):
            if 'else' in self._definition:
                self.generate_func_code_block(
                    self._definition['else'],
                    self._variable,
                    self._variable_name,
                    clear_variables=True
                )
            else:
                self.l('pass')
        if 'then' in self._definition:
            with self.l('else:'):
                self.generate_func_code_block(
                    self._definition['then'],
                    self._variable,
                    self._variable_name,
                    clear_variables=True
                )

    def generate_content_encoding(self):
        """
        Means decoding value when it's encoded by base64.

        .. code-block:: python

            {
                'contentEncoding': 'base64',
            }
        """
        if self._definition['contentEncoding'] == 'base64':
            with self.l('if isinstance({variable}, str):'):
                with self.l('try:'):
                    self.l('import base64')
                    self.l('{variable} = base64.b64decode({variable})')
                with self.l('except Exception:'):
                    self.l(
                        'raise JsonSchemaException("{name} must be encoded by base64")')
                with self.l('if {variable} == "":'):
                    self.l(
                        'raise JsonSchemaException("contentEncoding must be base64")')

    def generate_content_media_type(self):
        """
        Means loading value when it's specified as JSON.

        .. code-block:: python

            {
                'contentMediaType': 'application/json',
            }
        """
        if self._definition['contentMediaType'] == 'application/json':
            with self.l('if isinstance({variable}, bytes):'):
                with self.l('try:'):
                    self.l('{variable} = {variable}.decode("utf-8")')
                with self.l('except Exception:'):
                    self.l(
                        'raise JsonSchemaException("{name} must encoded by utf8")')
            with self.l('if isinstance({variable}, str):'):
                with self.l('try:'):
                    self.l('import json')
                    self.l('{variable} = json.loads({variable})')
                with self.l('except Exception:'):
                    self.l(
                        'raise JsonSchemaException("{name} must be valid JSON")')



# -*- coding: utf-8 -*-
# Copyright 2021 IRT Saint Exupéry, https://www.irt-saintexupery.com
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

# Contributors:
#    INITIAL AUTHORS - initial API and implementation and/or
#                      initial documentation
#        :author:  Francois Gallard
#    OTHER AUTHORS   - MACROSCOPIC CHANGES

"""Parse source code to extract information."""

from __future__ import division, unicode_literals

import inspect
import re
from typing import Callable, Dict

from gemseo.utils.py23_compat import getargspec


def get_options_doc(
    method,  # type: Callable
):  # type: (...) -> Dict[str, str]
    """Get the documentation of a method.

    Args:
        method: The method to retrieve the doc from.

    Returns:
        The descriptions of the options.
    """
    # the docstring has all the leading and common white spaces removed
    docstring = inspect.getdoc(method)

    if docstring is None:
        raise ValueError("Empty doc for {}".format(method))

    for parse in (parse_google, parse_rest):
        parsed_docstring = parse(docstring)
        if parsed_docstring:
            return parsed_docstring

    raise ValueError(
        "The docstring of the arguments is malformed: "
        "please use Google style docstrings"
    )


def get_default_options_values(
    cls,
):  # type: (...) -> Dict[str, str]
    """Get the options default values for the given class, by only addressing kwargs.

    Args:
        cls: The class.
    """
    args, _, _, defaults = getargspec(cls.__init__)
    if "self" in args:
        args.remove("self")
    n_def = len(defaults)

    return {args[-n_def:][i]: defaults[i] for i in range(n_def)}


def parse_rest(
    docstring,  # type: str
):  # type: (...) -> Dict[str, str]
    """Parse a reST docstring.

    Args:
        docstring: The docstring to be parsed.

    Returns:
        The parsed docstring.
    """
    pattern = r":param ([\*\w]+): (.*?)(?:(?=:param)|(?=:return)|\Z)"
    param_re = re.compile(pattern, re.S)
    doc_list = param_re.findall(docstring)
    parsed_doc = {txt[0]: txt[1].replace(" " * 4, "").rstrip("\n") for txt in doc_list}
    parsed_doc = {
        name: description.replace("  ", " ").replace("\n", " ").replace("  ", "\n\n")
        for name, description in parsed_doc.items()
    }
    return parsed_doc


# regex pattern for finding the arguments section of a google docstring
# doc_inherit.DocInheritMeta replaces the section title "Args" with "Parameters"
RE_PATTERN_ARGS_SECTION = re.compile(
    r"(?:Args|Parameters)\s*:\s*\n(.*?)(?:Returns:|Raises:|$)", flags=re.DOTALL
)

# regex pattern for finding the arguments names and description of a google docstring
RE_PATTERN_ARGS = re.compile(
    r"\**(\w+)\s*:\s*(.*?)(?:$|(?=\**\w+\s*:))", flags=re.DOTALL
)


def parse_google(
    docstring,  # type: str
):  # type: (...) -> Dict[str, str]
    """Parse a Google docstring.

    Args:
        docstring: The docstring to be parsed.

    Returns:
        The parsed docstring with the function arguments names bound to their descriptions.
    """
    args_sections = RE_PATTERN_ARGS_SECTION.findall(docstring)

    if len(args_sections) != 1:
        # This is not a google docstring
        return {}

    # remove leading common blank spaces
    args_section = inspect.cleandoc(args_sections[0])

    parsed_doc = {}

    for name, desc in RE_PATTERN_ARGS.findall(args_section):
        # remove multiple blank spaces and carriage returns
        parsed_doc[name] = re.sub(r"\s+", " ", desc).strip()

    return parsed_doc
